#ifndef HIT_TRACK_CHARGE_PROCESSOR_H
#define HIT_TRACK_CHARGE_PROCESSOR_H 1

//LCIO
#include <lcio.h>

//MARLIN
#include "marlin/Processor.h"

namespace AIDA{
  class IHistogram1D;
}


namespace marlintpc
{

    /** Fill charge per hit and per track length into AIDA histograms.
     *
     *  By default the chamber is divided into 10 z-bins and for each bin a 
     *  histogram is filled.<br>
     *  In addition one histogram is filled without any z cut.
     *  
     *  <h4>Input - Prerequisites</h4>
     *  \li A collection of reconstructed tracks <br>
     *      To have the charge only from hits on tracks,
     *      LCIO::TRBIT_HITS has to be set,
     *      so the hits are stored.
     *  \li A collection of hits
     *
     *  <h4>Output</h4>
     *  \li Histograms with charge per hit
     *  \li Histograms with charge per hit for hits in tracks only
     *  \li Histograms with charge per track
     *  \li Histograms with charge per track length (dEdx)
     *
     *  @param InputTracks Name of input Tracks collection (default: TPCTracks)
     *  @param InputHits Name of input TrackerHits collection (default: TPCHits)
     *  @param MaxChargePerHit  Maximum value in ChargePerHit histograms (default: ?? )
     *  @param MaxChargePerTrack Maximum value in ChargePerHit histogram (default: ?? )
     *  @param Max_dEdx  Maximum value in dEdx histogram (default: 1000 )
     *  @param MinZ Minimum z value of lowest bin in case of z binning (default: 0)
     *  @param MaxZ Maximum z value of highest bin in case of z binning (default: 260)
     *  @param nZBins Number ob bins in z (default: 10)<br>
     *         z binning switched of if 0, one histogram without z cut
     *
     *  All Histograms have 0 as lowest value
     *
     *  @author Martin Killenberg, University of Bonn
     */

    class HitAndTrackChargeProcessor : public marlin::Processor{
			      
    public:
    
	virtual Processor*  newProcessor() { return new HitAndTrackChargeProcessor ; }
    
	HitAndTrackChargeProcessor();

	~HitAndTrackChargeProcessor();

	virtual void init();

	virtual void processRunHeader( EVENT::LCRunHeader* run ) ;

	virtual void processEvent( LCEvent * evt ) ; 
    
    protected:

     std::string _tracksColName ;    ///< the name of the input tracks collection
     std::string _hitsColName ;    ///< the name of the input hits collection
     
     std::vector<AIDA::IHistogram1D *> _chargePerTrackHistos; ///< charge per track histos vector
     std::vector<AIDA::IHistogram1D *> _dEdxHistos; ///< dEdx of the track histos vector
     std::vector<AIDA::IHistogram1D *> _chargePerHitHistos; ///< charge per hit histos vector
     std::vector<AIDA::IHistogram1D *> _chargePerHitOnTrackHistos; ///< charge per hit for hits on tracks only histos vector

     float _maxChargePerHit; ///< Maximum value in ChargePerHit histograms
     float _maxChargePerTrack; ///< Maximum value in ChargePerHit histogram
     float _max_dEdx;  ///< Maximum value in dEdx histogram
     float _minZ; ///< Minimum z value of lowest bin
     float _maxZ; ///< Maximum z value of lowest bin
     
     int _nZBins; ///< Number of z bins
     
     float _zBinWidth;
 
    };
}

#endif // HIT_TRACK_CHARGE_PROCESSOR_H
