#include "HitAndTrackCounterProcessor.h"
#include "ZBinTemplateProcessor.cc"

//C++
#include <iostream>
#include <cmath>
using namespace std;

//LCIO
#include <IMPL/LCCollectionVec.h>
#include <EVENT/Track.h>

// Aida
//#ifdef MARLIN_USE_AIDA // the whole processor does not make sense without aida
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/IHistogram1D.h>
// endif

#include <marlin/Global.h>

using namespace lcio ;
using namespace marlin ;

namespace marlintpc
{

// define the indices for the template parameter
// they are for convenience and better code readability
const unsigned int HitAndTrackCounterProcessor::_hitsPerTrackHistosIndex   = 0;
//const unsigned int HitAndTrackCounterProcessor::_hitsPerEventHistosIndex   = 1;

HitAndTrackCounterProcessor aHitAndTrackCounterProcessor;
    
    HitAndTrackCounterProcessor::HitAndTrackCounterProcessor(): 
	ZBinTemplateProcessor<N_1D_HISTOS_HIT_TRACK_COUNTER_PROCESSOR ,
			      N_2D_HISTOS_HIT_TRACK_COUNTER_PROCESSOR>("HitAndTrackCounterProcessor"),
	_hitsPerEventHisto(0),
	_tracksPerEventHisto(0)
{
    // register steering parameters: name, description, class-variable, default value
    registerInputCollection(    LCIO::TRACK,
				"InputTracks" , 
				"Name of the input tracks collection (default: TPCTracks)"  ,
				_tracksColName ,
				std::string("TPCTracks") ) ;
    registerInputCollection(    LCIO::TRACKERHIT,
				"InputHits" , 
				"Name of the input hits collection (default: TPCHits)"  ,
				_hitsColName ,
				std::string("TPCHits") ) ;
    // modify processor description
    _description = "Count hits in events and tracks, and track per event" ;

    registerProcessorParameter( "MaxHitsPerTrack" , 
				"Maximum value in HitsPerTrack  histogram (default: 20 )"  ,
				_maxHitsPerTrack,
				int(20) ) ;

  registerProcessorParameter( "MaxHitsPerEvent" , 
			      "Maximum value in HitsPerEvent  histogram (default: 50 )"  ,
			      _maxHitsPerEvent,
			      int(50) ) ;

  registerProcessorParameter( "MaxTracksPerEvent" , 
			      "Maximum value in TracksPerEvent  histogram (default: 10 )"  ,
			      _maxTracksPerEvent ,
			      int(10) ) ;


  // set correct names for the z binned histograms
  _histo1DBaseNames[_hitsPerTrackHistosIndex]	   = "hitsPerTrackHisto";
  _histo1DTitles[_hitsPerTrackHistosIndex] = "Number of hits per track";
  
}

HitAndTrackCounterProcessor::~HitAndTrackCounterProcessor()
{
}

void HitAndTrackCounterProcessor::init()
{ 
	// usually a good idea to
	printParameters() ;

	// get the not-z-binned histograms from the AIDA processor
	_hitsPerEventHisto =   AIDAProcessor::histogramFactory(this)->
	  createHistogram1D( "hitsPerEventHisto", 
			     "Number of hits per event", 
			     _maxHitsPerEvent, 0, _maxHitsPerEvent);

	_tracksPerEventHisto =   AIDAProcessor::histogramFactory(this)->
	  createHistogram1D( "tracksPerEventHisto", 
			     "Number of tracks per event", 
			     _maxTracksPerEvent, 0, _maxTracksPerEvent);

	// set histo range and one bin per int for templated histograms
	// according to Processor parameters
	_histoRanges1D[_hitsPerTrackHistosIndex].max = _maxHitsPerTrack;
	_nBins1D[_hitsPerTrackHistosIndex] = _maxHitsPerTrack;

	// now call ZBinTemplateProcessor's init which initialises the histograms for us
	ZBinTemplateProcessor<N_1D_HISTOS_HIT_TRACK_COUNTER_PROCESSOR,
	                      N_2D_HISTOS_HIT_TRACK_COUNTER_PROCESSOR>::init();
}

void HitAndTrackCounterProcessor::processRunHeader( EVENT::LCRunHeader* run )
{
}
    
void HitAndTrackCounterProcessor::processEvent( LCEvent * evt ) 
{ 
    //read input data

    // we need a LCCollectionVec to use iterators, as LCCollectio is not a std::vector
    LCCollectionVec* tracksCollection=0;
    try{
	tracksCollection = dynamic_cast<LCCollectionVec*>(evt->getCollection( _tracksColName )) ;
    }
    catch(DataNotAvailableException &e)
    {
	// there are no tracks
	// nevermind, just catch the exception and continue 
	streamlog_out(DEBUG0) << "no input data returning"<<std::endl;
    }


    // if there are tracks count them and the hits per track
    if( tracksCollection != 0 )
    {
	for(LCCollectionVec::iterator it = tracksCollection->begin() ; 
	    it < tracksCollection->end(); it ++ )
	{
	    Track * track = dynamic_cast<Track *>(*it) ;
	    
	    TrackerHitVec hitVector = track->getTrackerHits();
	    float sumZ=0;
	
//	    cout << "nr of hits " << hitVector.size()<< endl;
	
//	    int hitNr=0;
	
//	    cout << "\t\t hit\t x \t y \t z \t residual_x/r*phi" << endl;
	    for (TrackerHitVec::iterator hit = hitVector.begin();
		 hit <  hitVector.end(); hit ++)
	    {
		const double *pos = (*hit)->getPosition();
		
		sumZ+=pos[2];
	}

	    
	    // fill the number if hits per track into the histogram, use z0 of the track as z position
	    fillHistogram( _hitsPerTrackHistosIndex, track->getTrackerHits().size() , sumZ/hitVector.size() );
	}

	_tracksPerEventHisto->fill(tracksCollection->getNumberOfElements());
    }
    else
    {
	_tracksPerEventHisto->fill( 0 );	
    }

    try{
	_hitsPerEventHisto->fill( evt->getCollection( _hitsColName )->getNumberOfElements() ) ;
    }
    catch(DataNotAvailableException &e)
    {
	// there are no hits
	// just continue
    }
}


}// namespace 
