#include "PerPadPulseChargeHistogrammerProcessor.h"

//C++
#include <iostream>
#include <cmath>
#include <sstream>

//LCIO
#include <IMPL/LCCollectionVec.h>
#include <EVENT/Track.h>
#include <EVENT/TrackerPulse.h>

// AIDA
//#ifdef MARLIN_USE_AIDA // the whole processor does not make sense without AIDA
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/IHistogram1D.h>
// endif

//MARLIN
#include <marlin/Global.h>

using namespace std;
using namespace lcio ;
using namespace marlin ;

namespace marlintpc
{

	PerPadPulseChargeHistogrammerProcessor aPerPadPulseChargeHistogrammerProcessor; ///< Necessary instance of the processor.

	PerPadPulseChargeHistogrammerProcessor::PerPadPulseChargeHistogrammerProcessor() : Processor ("PerPadPulseChargeHistogrammerProcessor")
	{
		// register steering parameters: name, description, class-variable, default value ...

		registerInputCollection (LCIO::TRACKERPULSE,
		                         "InputPulses",
		                         "Name of the input pulse collection (default: TPCPulses)",
		                         _pulsesColName,
		                         std::string ("TPCPulses"));

		registerProcessorParameter ("MinCharge",
		                            "Minimal charge value in histogram (default:  0)",
		                            _minCharge,
		                            float (0));

		registerProcessorParameter ("MaxCharge",
		                            "Maximal charge value in histogram (default:  300)",
		                            _maxCharge,
		                            float (300));

		registerProcessorParameter ("Bins",
		                            "Bins (default:  300)",
		                            _nBins,
		                            int (300));

		registerProcessorParameter ("Channels",
		                            "Number of Channels (default:  96)",
		                            _nChannels,
		                            int (96));

		// processor description
		_description = "Generates histograms of pulse charges for every channel";
	}

	PerPadPulseChargeHistogrammerProcessor::~PerPadPulseChargeHistogrammerProcessor()
	{
	}



	void PerPadPulseChargeHistogrammerProcessor::init()
	{
		// usually a good idea to
		printParameters() ;

		// resize vector to number of channels + 1 for the general histogram
		_chargePerPadHistos.resize (_nChannels + 1);

		// Create a histo for each channel and the last one to plot all pulses together ...
		// (due to special and general histograms being treated in the vector, adressing by index might be
		// more suitable than using iterator and asking position)
		for (int i = 0; i <= _nChannels; i++)
		{
			std::stringstream zRangeString;
			std::stringstream zIndexString;

			// Channel's histograms have the channel number attached ...
			if (i < _nChannels)
			{
				zRangeString << " Channel: " << i;
				zIndexString << i ;
			}

			// compose title and name of the histograms ...
			string histoTitle = string ("Charge per Pulse") + zRangeString.str();
			string histoName  = string ("ChargePerPulse") + zIndexString.str();

			streamlog_out (DEBUG0) << histoName << " : " << histoTitle << endl;

			// get a histogram from the AIDAProcessor
			_chargePerPadHistos[i] =   AIDAProcessor::histogramFactory (this)->
			                           createHistogram1D (histoName,
			                                              histoTitle,
			                                              _nBins, _minCharge, _maxCharge);
		}
	}



	void PerPadPulseChargeHistogrammerProcessor::processRunHeader (EVENT::LCRunHeader* run)
	{
		// Revision logging ...
		run->parameters().setValue (_processorName + "_revision", "$Rev: 1681 $");

		// Parameter logging ...
		for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
		{
			if (! i->second->isOptional() || i->second->valueSet())
			{
				run->parameters().setValue (_processorName + "_" + i->second->name(), i->second->value());
			}
		}
	}



	void PerPadPulseChargeHistogrammerProcessor::processEvent (LCEvent * evt)
	{
		//read input data ...
		streamlog_out (MESSAGE3) << "Event: " << evt->getEventNumber() << std::endl;

		LCCollectionVec* pulsesCollection = 0;
		
		try
		{
			pulsesCollection = dynamic_cast<LCCollectionVec*> (evt->getCollection (_pulsesColName));
		}
		catch (DataNotAvailableException &e)
		{
			// there are no pulses
			// continue with next event
			return;
		}

		// if the collection exists iterate the pulses to fill the charge into histograms ...
		if (pulsesCollection != 0)
		{
			for (LCCollectionVec::iterator pulseIter = pulsesCollection->begin();
			        pulseIter < pulsesCollection->end(); pulseIter ++)
			{
				TrackerPulse *pulse = dynamic_cast<TrackerPulse *> (*pulseIter);

				const int channel = pulse->getCellID0();

				// fill charge into histogram of channel ...
				if ( (channel >= 0) && (channel < _nChannels))
				{
					_chargePerPadHistos[channel]->fill (pulse->getCharge ());
				}

				// and once more into the general histogram
				_chargePerPadHistos[_nChannels]->fill (pulse->getCharge ());
			}
		}
	}
}// namespace
