#include "IonBackDriftProcessor.h"


//MARLINTPC
//#include "TPCVoxel.h"
#include "libgem.h"

//C++
#include <iostream>
#include <cmath>
#include <vector>

//LCIO
#include <IO/LCWriter.h>
#include <IMPL/LCEventImpl.h>
#include <IMPL/LCCollectionVec.h>

//GEAR
#include <gear/PadRowLayout2D.h>
#include <gear/TPCParameters.h>

///MARLIN
#include <marlin/Global.h>



using namespace lcio ;
using namespace marlin ;
using namespace std;


namespace marlintpc{

IonBackDriftProcessor aIonBackDriftProcessor;


IonBackDriftProcessor::IonBackDriftProcessor() : Processor("IonBackDriftProcessor"),_ionDisc(0) {

  _description = "IonBackDriftProcessor gives coordinates of ions escaping GEM structure.";


  //registers steering parameters, name, description, class-variable, default value

  registerInputCollection(    LCIO::LCGENERICOBJECT,
                              "InputCollectionName" ,
                              "Name of the input collection with amplified electrons (default: DistributedElectrons)",
                              _inputCollectionName,
                              std::string("DistributedElectrons") );
                          
  registerOutputCollection(   LCIO::LCGENERICOBJECT,
			      "OutputCollectionName",
                              "Name of the output collection with backdrifting ions (default: BackdriftingIons)",
                              _outputCollectionName,
                              std::string("BackdriftingIons") );

  registerProcessorParameter( "ParamFile" , 
			      "Name of the File containing Parametrisation of GEMs (default: parameterfile)"  ,
			      _paramFile ,
			      std::string("parameterfile") ) ;

  registerProcessorParameter( "OutputType" , 
			      "Determines whether output is a event snapshot or the whole disc (default: 0)"  ,
			      _outputType,
			      (int) 0) ;
  registerProcessorParameter( "ZBinLength" , 
			      "The length of the zBins of TPCVoxel (default: 2.5)"  ,
			      _zBinLength,
			      (double) 2.5) ;

  registerProcessorParameter( "Transient" , 
			      "switch off writing of output collection (default: 0)"  ,
			      _transient,
			      (int) 0) ;

  registerProcessorParameter( "UGEM1" , 
			      "Voltage on GEM 1 (default: 315)"  ,
			      _UGEM1 ,
			      (double)315) ;

  registerProcessorParameter( "UGEM2" , 
			      "Voltage on GEM 2 (default: 315)"  ,
			      _UGEM2 ,
			      (double)315) ;

  registerProcessorParameter( "UGEM3" , 
			      "Voltage on GEM 3 (default: 315)"  ,
			      _UGEM3 ,
			      (double)315) ;

  registerProcessorParameter( "ETrans1" , 
			      "Value of Transfer Field 1 (default: 2500)"  ,
			      _ETrans1 ,
			      (double)2500) ;

  registerProcessorParameter( "ETrans2" , 
			      "Value of Transfer Field 2 (default: 2500)"  ,
			      _ETrans2 ,
			      (double)2500) ;

  registerProcessorParameter( "EInd" , 
			      "Value of Induction Field (default: 5000)"  ,
			      _EInd ,
			      (double)5000) ;

  registerProcessorParameter( "DriftField",
                              "Value of the Drift Field (default: 240)" ,
			      _E ,
                              (double) 240);

}

void IonBackDriftProcessor::init() {

  streamlog_out(MESSAGE2) <<_description << endl;

  //Always a good idea....
  printParameters();

  _nRun = 0 ;
  _nEvt = 0 ;

  //Get the pad layout from GEAR.
  _padLayout = &(Global::GEAR->getTPCParameters().getPadLayout());

  //VoxelTPC for the build up of the ion disc.
  if( _ionDisc )
    delete _ionDisc;
  _ionDisc = new marlintpc::VoxelTPC( _padLayout, _zBinLength);

  //Calculate IonBackDrift from GEM settings.
  LoadGEMSettings();
}


void IonBackDriftProcessor::processRunHeader( LCRunHeader* run) { 

  run->parameters().setValue( _processorName + "_revision", "$Rev: 1437 $" );

  for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
    if ( ! i->second->isOptional() || i->second->valueSet() ) {
      run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
    }
  }

  _nRun++ ;
} 


void IonBackDriftProcessor::processEvent( LCEvent * evt ) {

  LCCollection* inputCol = 0;

  int entries = 1;

  try{
    inputCol = evt->getCollection( _inputCollectionName ) ;
    entries=inputCol->getNumberOfElements();
  }catch(EventException &exception){
    streamlog_out(DEBUG3) << exception.what() << endl;
    entries=0;
  }


  //Add backdrifting ions to voxels.
  for(int i=0;i<entries;i++) {
    TPCVoxel* tempVoxel = new TPCVoxel(dynamic_cast<LCGenericObject*>(inputCol->getElementAt(i)));
    int charge=static_cast<int>(fabs(tempVoxel->getCharge()*_ionBackDrift));
    _ionDisc->putChargeInVoxel( tempVoxel->getPadIndex(), 1, charge );
    delete tempVoxel;
  }

  LCCollectionVec* outputCol = 0;

  //Set output type: 0 for eventbased, other for total 
  if( _outputType )
    outputCol = _ionDisc->getVoxelCollection();
  else
    outputCol = _ionDisc->fetchVoxelCollection();

  if ( !outputCol ){
    return;
  }

  //Every voxel collection needs to contain the length of the zBins.
  float voxelZBinLength=static_cast<float>(_zBinLength);
  outputCol->parameters().setValue("zBinLength",voxelZBinLength);

  //To store the value of the ionbackdrift in output collection conversion to float is necessary.
  float backDrift=_ionBackDrift;
  outputCol->parameters().setValue( "IonBackDrift" , backDrift );

  //Set transient.
  if( _transient )
    outputCol->setTransient( true );

  //Write output collection.
  if( outputCol->empty() && _nEvt!=0 ) 
    delete outputCol;
  else //Adds output collection to the event.
    evt->addCollection( outputCol , _outputCollectionName );
  
  //if(_nEvt%100==0)
    streamlog_out(MESSAGE3) <<"        Event "<< evt->getEventNumber() <<" done!"<<endl;

  _nEvt ++ ;
}


//loads GEM settings from conditions data file
void IonBackDriftProcessor::LoadGEMSettings() {

  char *infile = const_cast<char*>(_paramFile.c_str());
  if( ! readpar( infile ) ) {
    streamlog_out(WARNING2) << "Not all parameters could be read from parameter file " 
                            << _paramFile << "!" << endl << "Exiting.";
    exit(1);
  }
  else 
    streamlog_out(DEBUG4) << "Parameter file read successfully " << endl;

  double params[7] = {_E, _UGEM1, _ETrans1, _UGEM2, _ETrans2, _UGEM3, _EInd}; 
  _ionBackDrift = ionf( params );

  streamlog_out(MESSAGE0) <<"Settings: "<< _E <<" "<< _UGEM1<<" "<< _ETrans1
                          <<" "<< _UGEM2 <<" "<< _ETrans2<<" "<< _UGEM3<<" "
                          << _EInd << endl << "Ionbackdrift: "<< _ionBackDrift << endl;

}


}//end of namespace marlintpc
