#include <algorithm>
//#include <vector>
//#include <string>
#include <cmath>
#include <iostream>
#include <iomanip>
#include <Exceptions.h>

#include <IMPL/TrackerRawDataImpl.h> 

#include "TPCElectronicsProcessor.h"

namespace marlintpc
{

TPCElectronicsProcessor::ADCPulse::ADCPulse(double frequency, double riseTime, double nElectronsDynamicRange, 
		   int ADCMaxValue)
  :  _frequency( frequency/1000.), // internally use frequency in GHz (avoid many multiplications with 1000)
     _riseTime(riseTime), _ADCMaxValue(ADCMaxValue),
     _sqrt2_sigma( riseTime/3. *sqrt(2) ), _startBin(-1),
     _channelNumber(-1)
{
  // Calculate the maximal possible charge in one time bin originating from 
  // nElectronsDynamicRange electrons.
  // This value should correspond to ADCMaxValue.

  // The maximum charge is in one bin if the centre of the bin is the centre of the gaussian.
  // So the bin from -halfTimeBin to +halfTimeBin (if the centre is 0 like in the standard erf).
  double halfTimeBin = 0.5/_frequency;
  double maxPossibleChargeIntegral = nElectronsDynamicRange /2. *
                                     (   erf(  halfTimeBin / _sqrt2_sigma) 
				       - erf( -halfTimeBin / _sqrt2_sigma) );
  _ADCChargeFactor = _ADCMaxValue / maxPossibleChargeIntegral;
}

double TPCElectronicsProcessor::ADCPulse::getEndTime() const
{
  return (_startBin+_rawData.size()+1)/_frequency ;
}

int TPCElectronicsProcessor::ADCPulse::getChannelNumber() const
{
  return _channelNumber;
}

void TPCElectronicsProcessor::ADCPulse::clear(int newChannelNumber)
{
  _rawData.clear();
  _channelNumber=newChannelNumber;
  _startBin=-1;
}

void TPCElectronicsProcessor::ADCPulse::add(double startTime, int nElectrons)
{
  // Check that the pulses really overlap
  if ( (startTime > getEndTime()) && (_startBin >= 0) )
  {
    //    throw Exception("ADCPulse: ADCPulses do not overlap. Check the calling code!") ;
    throw Exception("ADCPulse: ADCPulses do not overlap. Check the calling code!") ; 
  }

  // Check that the start time of the new charge is not before the start time of the existing pulse.
  // This cannot be handles by this algorithm, the calling code has to take care of this.
  if ( startTime < _startBin/_frequency )
  {
    throw Exception("ADCPulse: Error: Start time of new charge is before the start of the existing pulse. Check the calling code!") ; 
  }  

  // the end time is start time + the width of the ADCPulse, currently 2*risetime
  double endTime = startTime + 2 * _riseTime;
  
  // the centre fo the Gaussian is start time plus rise time
  double centreTime = startTime + _riseTime;

  // the ADC values are calculated by intergating the gaussian inside the time bin and multiplying
  // with the ADCChargeFactor
  
  // start integration at start time for the first bin, the following bins start at the edge of the bin
  double integrationStartTime =  startTime;

  int bin = static_cast<int>(startTime* _frequency); // bin number fo the first bin with charge
  if (_startBin<0) // if there is no charge in the pulse yet the pulse starts at "bin"
  {
    _startBin = bin;
  }

  // The stop time fo the integration is the end of the bin.
  // If the end time is within the bin only integrate up to this point
  double integrationStopTime = std::min(endTime, (bin+1)/_frequency);

  
  double ADCChargeValue = nElectrons * _ADCChargeFactor;

  do{
    // calculate the ADC value for this bin by integrating the gaussian from 
    // integrationStartTime to integrationStopTime with a centre of the gaussion at centreTime

    double ADCValue =ADCChargeValue/2.*( erf( (integrationStopTime  - centreTime)/_sqrt2_sigma) 
				      -erf( (integrationStartTime - centreTime)/_sqrt2_sigma) );    
    
    // debug: check for 0 
    if (ADCValue==0)
    {
      streamlog_out(DEBUG3) << "integration from " <<integrationStartTime
			    << " to " << integrationStopTime
			    << " resulted in zero" << std::endl;
    }

    // check if the bin already exists in the vector, else append it
    if( (bin - _startBin) < static_cast<int>( _rawData.size()) )
    {
      // time bin is already there, add the charge
      _rawData[bin-_startBin] += ADCValue;
    }
    else // the time bin did not exist yet
    {
      _rawData.push_back( ADCValue );
    }

    // this is done later to get the full info when debugging
//   // perform ADC out of range check
 //   if (_rawData[bin-_startBin] > _ADCMaxValue)
 //   {
 //     _rawData[bin-_startBin] = _ADCMaxValue;
 //   } 
    
    // move to the next time bin
    integrationStartTime = integrationStopTime;
    bin++;
    integrationStopTime = std::min(endTime, (bin+1)/_frequency);
    

  }while(integrationStartTime < endTime);

}

const std::vector<float> & TPCElectronicsProcessor::ADCPulse::getRawDataVec() const
{
  return _rawData;
}

  EVENT::TrackerRawData * TPCElectronicsProcessor::ADCPulse::getLCIORawData() const
 {
   std::vector<float>::const_iterator StartOfPulse= _rawData.begin();
   std::vector<float>::const_iterator EndOfPulse= _rawData.end();
   int lcioStartBin = _startBin;

   while ( (StartOfPulse < _rawData.end()) && (*StartOfPulse < 0.5) )
   {
     streamlog_out(DEBUG1) << "startsignal " << std::scientific<< *StartOfPulse ;

     StartOfPulse++;
     lcioStartBin++;

     streamlog_out(DEBUG1) << " ; startBin now is "<< lcioStartBin  << std::endl;

   }

   while ( (EndOfPulse > _rawData.begin() ) && (*EndOfPulse < 0.5) )
   {
     streamlog_out(DEBUG1) << "endsignal " << std::scientific<< *EndOfPulse << std::endl ;
     EndOfPulse--;
   }
   
   if (EndOfPulse > StartOfPulse)
   {
     std::vector<short> shortRawDataVec( std::distance(StartOfPulse, EndOfPulse) );

     for ( std::vector<float>::const_iterator binIter = StartOfPulse; 
	   binIter < EndOfPulse; binIter++ )
     {
       if ( *binIter < _ADCMaxValue )
       {
	 shortRawDataVec[ std::distance(StartOfPulse, binIter) ] = static_cast<short>(round(*binIter));
       }
       else
       {
	 streamlog_out(DEBUG3) << "pulse on channel " <<_channelNumber
			       << " out of range (" << *binIter <<"/"<< _ADCMaxValue
			       << " in bin " << std::distance(StartOfPulse, binIter)
			       << std::endl;
	 shortRawDataVec[ std::distance(StartOfPulse, binIter) ] =  _ADCMaxValue;
       }
     }
     
     IMPL::TrackerRawDataImpl *lcioRawData = new IMPL::TrackerRawDataImpl;
     lcioRawData->setADCValues( shortRawDataVec );
     lcioRawData->setCellID0( _channelNumber );
     lcioRawData->setTime( lcioStartBin );

     return lcioRawData;
   }
   else
   {
     streamlog_out(DEBUG3) << "Empty pulse on channel " << _channelNumber;
     streamlog_out(DEBUG2) << " ; Charge values ";
     if( streamlog::out.write< streamlog::DEBUG3 >() )
     {
       double sum=0;
       for (std::vector<float>::const_iterator binIter= _rawData.begin();
	    binIter < _rawData.end(); binIter++)
       {
	 streamlog_out(DEBUG2) << *binIter << " : ";
	 sum += *binIter;
       }
       streamlog_out(DEBUG3) << " SumQ " << sum << std::endl;
     }
     return 0;
   }
 }

}// namespace marlintpc
