#ifdef USE_LCCD

// MarlinTPC
#include "PhotoelectricLikelihoodFittingProcessor.h"

// Marlin
#include <marlin/Exceptions.h>
#include <marlin/Global.h>

// GEAR
#include <gear/TPCModule.h>
#include <gear/TPCParameters.h>
#include <gear/PadRowLayout2D.h>
#include <gearxml/tinyxml.h>

// LCIO
#include <EVENT/LCIO.h>
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerPulse.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerHitImpl.h>

// AIDA
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/ITupleFactory.h>
#include <AIDA/ITuple.h>

// Minuit
#include <Minuit2/MnPrint.h>
#include <Minuit2/MnUserParameters.h>
#include <Minuit2/MnUserCovariance.h>
#include <Minuit2/MnMinimize.h>
#include <Minuit2/FunctionMinimum.h>

using namespace CLHEP;
using namespace EVENT;
using namespace lcio;
using namespace gear;
using namespace marlin;
using namespace ROOT;
using namespace Minuit2;
using namespace std;

namespace marlintpc {


PhotoelectricLikelihoodFittingProcessor aPhotoelectricLikelihoodFittingProcessor;


PhotoelectricLikelihoodFittingProcessor::PhotoelectricLikelihoodFittingProcessor() : Processor("PhotoelectricLikelihoodFittingProcessor"),

	_input_seed_photodot_collection_name ( "TPCSeedPhotodots" ),
	_input_seed_photoline_collection_name ( "TPCSeedPhotolines" ),
	_output_photodot_collection_name ( "TPCPhotodots" ),
	_output_photoline_collection_name ( "TPCPhotolines" ),
	_output_is_transient (true),
	_sigma ( 2.5 ) {

	/* description of the processor which will be displayed in the
	* steering file autometicly generated by
	* marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
	*/

	_description = "PhotoelectricLikelihoodFittingProcessor performs a likelihood fit on photoelectric data.";

	/* register steering parameters: name, description, class-variable, default value
	* the type will by definded by the type of the default value
	* string, double, float and int are possible
	* plaese note to get a float you must cast in C++: (float) 1.3
	*/

	registerInputCollection( LCIO::TRACKERHIT,
		"InputSeedPhotodots",
		"The name of the input collection of TrackerHits which represent seed photodots.",
		_input_seed_photodot_collection_name,
		_input_seed_photodot_collection_name );

	registerInputCollection( LCIO::TRACK ,
		"InputSeedPhotolines",
		"The name of the input collection of Tracks which represent seed photolines.",
		_input_seed_photoline_collection_name,
		_input_seed_photoline_collection_name );

	registerOutputCollection( LCIO::TRACKERHIT,
		"OutputPhotodots",
		"The name of the output collection of TPCHits which represent fitted photodots.",
		_output_photodot_collection_name,
		_output_photodot_collection_name );

	registerOutputCollection( LCIO::TRACK,
		"OutputPhotolines",
		"The name of the output collection of TPCHits which represent fitted photolines.",
		_output_photoline_collection_name,
		_output_photoline_collection_name );

	registerOptionalParameter( "SetOutputTransient",
		"Controls whether the output collection should be saved in the LCIO file (false) or if it should be deleted at the end (true - default).",
		_output_is_transient,
		true );

	registerOptionalParameter( "Sigma",
		"The expected diffusion of photodot charge due to the drift distance and initial defocussing.",
		_sigma,
		_sigma );

}


void PhotoelectricLikelihoodFittingProcessor::init() {

	// usually a good idea to
	printParameters();

	// load AIDA (if it's available)
	try {

		_photodot_fit_results_tuple = AIDAProcessor::tupleFactory( this )->create( "PhotodotFitResults", "Photodot Fit Results", "int i_dot, float x, float y, int n_pads, bool fit_status","" );

	}
	catch ( lcio::Exception &e ) {

		streamlog_out( DEBUG ) << e.what() << endl;
		return;
	}

}


void PhotoelectricLikelihoodFittingProcessor::processRunHeader( LCRunHeader* run) {

}


void PhotoelectricLikelihoodFittingProcessor::processEvent( LCEvent * evt ) {

	// the input seed photodots
	LCCollection *seed_photodot_collection = evt->getCollection ( _input_seed_photodot_collection_name );
	const int seed_photodot_count = seed_photodot_collection->getNumberOfElements();

	// the output fit photodots
	LCCollectionVec* output_fit_photodots = new LCCollectionVec ( LCIO::TRACKERHIT );
	output_fit_photodots->setTransient ( _output_is_transient );
	output_fit_photodots->setSubset ( false );

	// loop through the seed photodots, fitting them
	for ( int i_dot = 0; i_dot < seed_photodot_count; i_dot++ ) {

		TrackerHit *seed_photodot = dynamic_cast< TrackerHit* > ( seed_photodot_collection->getElementAt ( i_dot ) );

		assert ( seed_photodot );

		// create a TrackerHit to store the results
		TrackerHitImpl *fit_photodot = new TrackerHitImpl;
		output_fit_photodots->addElement ( fit_photodot );

		// check to see if any fitting needs to occur
		if ( seed_photodot->getRawHits().size() == 0 ) {

			streamlog_out ( DEBUG ) << "Skipping photodot " << i_dot << ". It didn't receive any data this event." << endl;
			_photodot_fit_results_tuple->fill( 0, i_dot );
			_photodot_fit_results_tuple->fill( 3, 0 );
			_photodot_fit_results_tuple->fill( 4, false );
			_photodot_fit_results_tuple->addRow();
			continue;
		}

		const double *seed_position = seed_photodot->getPosition();

		if ( seed_photodot->getRawHits().size() == 1 ) {

			streamlog_out ( DEBUG ) << "Skipping photodot " << i_dot << ". It only received data from one pad this event." << endl;

			// set the position
			double position[3];
			position[0] = seed_position[0];
			position[1] = seed_position[1];
			position[2] = seed_position[2];
			fit_photodot->setPosition ( position );

			// set the covariance
			float covariance[6];
			covariance[0] = 0.;
			covariance[1] = 0.;
			covariance[2] = 0.;
			covariance[3] = 0.;
			covariance[4] = 0.;
			covariance[5] = 0.;
			fit_photodot->setCovMatrix ( covariance );

			if ( _photodot_fit_results_tuple ) {

				_photodot_fit_results_tuple->fill( 0, i_dot );
				_photodot_fit_results_tuple->fill( 1, position[0] );
				_photodot_fit_results_tuple->fill( 2, position[1] );
				_photodot_fit_results_tuple->fill( 3, 1 );
				_photodot_fit_results_tuple->fill( 4, true );
				_photodot_fit_results_tuple->addRow();
			}

			continue;
		}

		streamlog_out( DEBUG ) << "Fitting photodot " << i_dot << " with seed position (" << seed_position[0] << ", " << seed_position[1] << ")." << endl;

		// create the objects for doing the fitting
		PhotodotLikelihood photodot_likelihood( *seed_photodot );

		MnUserParameters user_parameters;

		user_parameters.Add( "x", seed_position[0], 1.E-01 );
		user_parameters.Add( "y", seed_position[1], 1.E-01 );
		user_parameters.Add( "sigma", _sigma );

		MnMinimize minimizer( photodot_likelihood, user_parameters );

		FunctionMinimum minimum = minimizer();

		// store the reconstructed position
		const MnUserParameters &fit_parameters = minimum.UserParameters();
		double fit_position[3];
		fit_position[0] = fit_parameters.Value ( "x" );
		fit_position[1] = fit_parameters.Value ( "y" );
		fit_position[2] = 0.;
		fit_photodot->setPosition ( fit_position );

		if ( minimum.HasCovariance() ) {
			// store the covariance matrix
			// cov(x,x) , cov(y,x) , cov(y,y) , cov(z,x) , cov(z,y) , cov(z,z)
			const MnUserCovariance &fit_covariance = minimum.UserCovariance();
			float covariance[6];
			covariance[0] = fit_covariance ( 0, 0 );
			covariance[1] = fit_covariance ( 0, 1 );
			covariance[2] = fit_covariance ( 1, 1 );
			covariance[3] = 0.;
			covariance[4] = 0.;
			covariance[5] = 0.;
			fit_photodot->setCovMatrix ( covariance );
		}

		streamlog_out ( DEBUG ) << minimum;

		_photodot_fit_results_tuple->fill( 0, i_dot );
		_photodot_fit_results_tuple->fill( 1, fit_parameters.Value( "x" ) );
		_photodot_fit_results_tuple->fill( 2, fit_parameters.Value( "y" ) );
		_photodot_fit_results_tuple->fill( 3, (int) seed_photodot->getRawHits().size() );
		_photodot_fit_results_tuple->fill( 4, minimum.IsValid() );
		_photodot_fit_results_tuple->addRow();
	}

	evt->addCollection ( output_fit_photodots, _output_photodot_collection_name );
}



void PhotoelectricLikelihoodFittingProcessor::check( LCEvent * evt ) {

	// nothing to check here - could be used to fill checkplots in reconstruction processor
}


void PhotoelectricLikelihoodFittingProcessor::end() {

}


PhotoelectricLikelihoodFittingProcessor::PhotodotLikelihood::PhotodotLikelihood( const TrackerHit &photodot_data ):
			_photodot_data( photodot_data ) {

}


double PhotoelectricLikelihoodFittingProcessor::PhotodotLikelihood::operator()( const vector<double>& x ) const {

	// the parameters are global x, global y, and sigma
	const double sigma = x[2];

	// get the TPC parameters
	const gear::TPCParameters &tpc_parameters = Global::GEAR->getTPCParameters();

	double negative_log_likelihood = 0.;

	float c0, c1;
	// convert the parameters to the global coordinate system of the TPC
	if ( tpc_parameters.getCoordinateType() == PadRowLayout2D::CARTESIAN ) {

		c0 = x[0];
		c1 = x[1];
	}
	else if ( tpc_parameters.getCoordinateType() == PadRowLayout2D::POLAR ) {

		c0 = sqrt( x[0]*x[0] + x[1]*x[1] );
		c1 = atan2( x[1], x[0] );
	}

	// iterate through the pads calculating how much charge they were expected to collect
	const LCObjectVec &pulses = _photodot_data.getRawHits();

	double total_charge = 0.;

	for( LCObjectVec::const_iterator i_data_pulse = pulses.begin(); i_data_pulse != pulses.end(); i_data_pulse++ ) {

		const TrackerPulse *pulse = dynamic_cast< TrackerPulse* > (*i_data_pulse);

		const TPCModule &module = tpc_parameters.getModule( pulse->getCellID1() );

		Vector2D local_photodot_coords = module.globalToLocal( c0, c1 );

		Vector2D global_pad_centre = module.getPadCenter( pulse->getCellID0() );

		Vector2D local_pad_centre = module.globalToLocal( global_pad_centre[0], global_pad_centre[1] );

		double vx;		// the centre of the dot, relative to the pad centre
		double vy;		// the centre of the dot, relative to the pad centre

		double half_pad_width = 0.5 * module.getPadWidth( pulse->getCellID0() );
		double half_pad_height = 0.5 * module.getPadHeight( pulse->getCellID0() );

		// convert the result to cartesian coordinates
		switch ( module.getLocalPadLayout().getCoordinateType() ) {

			case PadRowLayout2D::POLAR: {

				float r = local_photodot_coords[0];
				local_photodot_coords[0] = r * cos( local_photodot_coords[1] );
				local_photodot_coords[1] = r * sin( local_photodot_coords[1] );

				r = local_pad_centre[0];
				half_pad_width *= r;
				const double sin_p = sin( local_pad_centre[1] );
				const double cos_p = cos( local_pad_centre[1] );

				local_pad_centre[0] = r * cos_p;
				local_pad_centre[1] = r * sin_p;

				// rotate the displacement vector so the pad height is aligned with the local y-axis
				const double vxp = local_photodot_coords[0] - local_pad_centre[0];
				const double vyp = local_photodot_coords[1] - local_pad_centre[1];

				vx = (vxp * sin_p) - (vyp * cos_p);
				vy = (vyp * sin_p) + (vxp * cos_p);
			}
			break;

			case PadRowLayout2D::CARTESIAN: {

				vx = local_photodot_coords[0] - local_pad_centre[0];
				vy = local_photodot_coords[1] - local_pad_centre[1];

			}
			break;

			default: throw gear::Exception("bad layout type");

		}


		// now perform the integral
		const double d = M_SQRT1_2 / sigma;
		const double A = erf( (vx + half_pad_width)*d  ) - erf( (vx - half_pad_width)*d  );
		const double B = erf( (vy + half_pad_height)*d ) - erf( (vy - half_pad_height)*d );
		const double expected_charge = 0.25 * A * B;

		negative_log_likelihood -= std::log( expected_charge ) * pulse->getCharge();

		total_charge += pulse->getCharge();
	}

	return negative_log_likelihood;
}


double PhotoelectricLikelihoodFittingProcessor::PhotodotLikelihood::Up() const {

	return 0.5;
}


}

#endif // USE_LCCD
