// MarlinTPC
#include "PhotoelectricPulseExtractionProcessor.h"

// Marlin
#include <marlin/Exceptions.h>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerPulse.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerPulseImpl.h>
#include <IMPL/LCFlagImpl.h>

// GEAR
#include <gearxml/tinyxml.h>

using namespace lcio;
using namespace marlin;
using namespace std;
using namespace gear;

namespace marlintpc {

PhotoelectricPulseExtractionProcessor aPhotoelectricPulseExtractionProcessor;


PhotoelectricPulseExtractionProcessor::PhotoelectricPulseExtractionProcessor() : Processor("PhotoelectricPulseExtractionProcessor"),

	_input_tracker_pulses_collection_name("TPCPulses"),
	_output_photoelectric_tracker_pulses_collection_name("TPCPhotoelectricPulses"),
	_output_is_transient(true),
	_projected_photogeometry_file_name("projected_photogeometry.xml") {

	/* description of the processor which will be displayed in the
	* steering file autometicly generated by
	* marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
	*/

	_description = "PhotoelectricPulseExtractionProcessor searches for pulses from the photoelectric system and puts then into a subset collection.";

	/* register steering parameters: name, description, class-variable, default value
	* the type will by definded by the type of the default value
	* string, double, float and int are possible
	* plaese note to get a float you must cast in C++: (float) 1.3
	*/
	registerInputCollection( LCIO::TRACKERPULSE ,
			"InputTrackerPulses",
			"Name of the input collection as stored in the LCIO file",
			_input_tracker_pulses_collection_name,
			std::string("TPCPulses") );

	registerOutputCollection( LCIO::TRACKERPULSE ,
			"OutputPhotoelectricTrackerPulses",
			"The name of the output collection of tracker pulses from the photoelectric system.",
			_output_photoelectric_tracker_pulses_collection_name,
			std::string("TPCPhotoelectricPulses") );

	registerOptionalParameter( "SetOutputTransient",
			"if not 0 the output collection is set transient",
			_output_is_transient,
			true);

	registerOptionalParameter( "ProjectedPhotogeometryFilename",
			"This is the name of the collection which stores the true positions of the photodots as SimTrackerHit objects. It is optional but it provides the number of photodots and, if provided, the processor will order the seed photodots to match the order of true photodots (useful for analysis).",
			_projected_photogeometry_file_name,
			_projected_photogeometry_file_name );

}


void PhotoelectricPulseExtractionProcessor::init() {

	// usually a good idea to
	printParameters();

	// open the projection file and get the time interval we're watching for and the list of pads to take
	TiXmlDocument document ( _projected_photogeometry_file_name.c_str() );

	if( !document.LoadFile() ) {

		streamlog_out( ERROR ) << "Couldn't load projected geometry file " << _projected_photogeometry_file_name << "." << endl;
		throw ( SkipEventException ( this ) );
	}

	// get the arrival time node
	TiXmlElement *arrival_element = document.RootElement()->FirstChildElement("arrival_time");

	if ( ! arrival_element ) {

		throw DataNotAvailableException ("Projected geometry file doesn't contain an <arrival_time> element.");
	}

	arrival_element->QueryDoubleAttribute( "start", &_photopulse_interval_start );
	arrival_element->QueryDoubleAttribute( "end", &_photopulse_interval_end );

	streamlog_out( DEBUG ) << "Photopulse interval is from " << _photopulse_interval_start << " to " << _photopulse_interval_end << " ns." << endl;

	// get a list of global pad indices to search for
	TiXmlElement *photodot_element = document.RootElement()->FirstChildElement ( "photodot" );

	for ( ; photodot_element; photodot_element = photodot_element->NextSiblingElement ( "photodot" ) ) {

		TiXmlElement *projection_element = photodot_element->FirstChildElement( "projection" );

		if ( ! projection_element ) {

			streamlog_out( DEBUG ) << "Found a phptodot with no projection." << endl;
			continue;
		}

		TiXmlElement *pad_element = projection_element->FirstChildElement ( "pad" );

		for ( ; pad_element; pad_element = pad_element->NextSiblingElement ( "pad" ) ) {

			int pad_index;
			int module_id;

			pad_element->QueryIntAttribute( "pad_index", &pad_index );
			pad_element->QueryIntAttribute( "module_id", &module_id );

			_global_pad_indices.insert( GlobalPadIndex( pad_index, module_id ) );
			streamlog_out( DEBUG ) << "Valid pad index " << pad_index << " module " << module_id << "." << endl;
		}
	}

	TiXmlNode *photoline_node = document.RootElement()->FirstChild("photoline");

	for ( ; photoline_node; photoline_node = photoline_node->NextSibling( "photoline" ) ) {

		TiXmlElement *projection_element = photoline_node->FirstChildElement( "projection" );

		if ( ! projection_element ) {

			streamlog_out( DEBUG ) << "Found a phptodot with no projection." << endl;
			continue;
		}

		TiXmlElement *pad_element = photoline_node->FirstChildElement( "pad" );

		for ( ; pad_element; pad_element = pad_element->NextSiblingElement( "pad" ) ) {

			int pad_index;
			int module_id;

			pad_element->QueryIntAttribute( "index", &pad_index );
			pad_element->QueryIntAttribute( "module", &module_id );

			_global_pad_indices.insert( GlobalPadIndex( pad_index, module_id ) );
			streamlog_out( DEBUG ) << "Valid pad index " << pad_index << " module " << module_id << "." << endl;
		}
	}
}


void PhotoelectricPulseExtractionProcessor::processRunHeader( LCRunHeader* run) {

}

void PhotoelectricPulseExtractionProcessor::processEvent( LCEvent * evt ) {


	LCCollection* input_pulses = evt->getCollection( _input_tracker_pulses_collection_name );

	streamlog_out( MESSAGE ) << "Searching " << input_pulses->getNumberOfElements() << " pulses from the collection " << _input_tracker_pulses_collection_name << endl;

	// create the output collection with a string defining the TYPE (see lcio.h)
	LCCollectionVec* output_photoelectric_pulses = new LCCollectionVec( LCIO::TRACKERPULSE );
	// set the transient flag as definded by the processor parameter
	output_photoelectric_pulses->setTransient( _output_is_transient );
	output_photoelectric_pulses->setSubset( true );

	// set the flags that cellID1 should be stored
	/*lcio::LCFlagImpl trkFlag(0) ;
	trkFlag.setBit( lcio::LCIO::TRAWBIT_ID1 ) ;
	output_photoelectric_pulses->setFlag( trkFlag.getFlag() );*/

	// now determine if this is a photoelectric pulse
	for ( int i_tracker_pulse = 0; i_tracker_pulse < input_pulses->getNumberOfElements(); i_tracker_pulse++ ) {

		TrackerPulse* input_pulse = dynamic_cast<TrackerPulse*>( input_pulses->getElementAt( i_tracker_pulse ) ) ;

		if ( !input_pulse ) {

			throw( Exception("Input collection " + _input_tracker_pulses_collection_name + " does not contain TrackerPulses.") );
		}

		const float time = input_pulse->getTime();

		// check if it's in the required time interval
		if ( ( time >= _photopulse_interval_start ) && ( time <= _photopulse_interval_end ) ) {

			// determine if it is one of the required pads
			const GlobalPadIndex global_index( input_pulse->getCellID0(), input_pulse->getCellID1() );

			if ( _global_pad_indices.find( global_index ) != _global_pad_indices.end() ) {

				output_photoelectric_pulses->addElement( input_pulse );
				streamlog_out( DEBUG ) << "Extracted pad " << global_index.getPadIndex() << " from module " << global_index.getModuleID() << " with charge " << input_pulse->getCharge() << " for the photoelectric system." << endl;
			}
		}
		else {

			streamlog_out ( DEBUG ) << "Pulse rejected. Reason? time: " << time << "." << endl;
		}
	}

	streamlog_out( MESSAGE ) << "Found " << output_photoelectric_pulses->getNumberOfElements() << " pulses which could be from the photoelectric system." << endl;

	evt->addCollection( output_photoelectric_pulses, _output_photoelectric_tracker_pulses_collection_name );
}



void PhotoelectricPulseExtractionProcessor::check( LCEvent * evt ) {
	// nothing to check here - could be used to fill checkplots in reconstruction processor
}


void PhotoelectricPulseExtractionProcessor::end() {

}

}
