#ifndef MARLINTPC_HIT_SPLITTER_H
#define MARLINTPC_HIT_SPLITTER_H

//stl
#include <set>

// lcio
#include <IMPL/LCCollectionVec.h>
#include <EVENT/TrackerHit.h>
#include <EVENT/TrackerPulse.h>
#include <EVENT/LCIO.h>

namespace marlintpc
{
  /** The hit splitter separates hits that are merged.
   *
   *  In a first step it looks for multiple pulses per channel and regroups the
   *  hit candidates so there is only one pulse per channel. It always associates the
   *  pulse on the neighbouring channels that are closest in time, if they overlap.
   *
   *  In a second step it searches for minima on a per pad basis, using the SignalSplitter class  
   *  to separate overapping signals.
   */

  class HitSplitter
  {
    public:
    /**  The constructor needs two collection it can write to: One for the split hits
     *   and one for the pulses that are contained in these hists (pulses might be split
     *   between hits, so they are not necessarily the same as the input pulses).
     *   In addition it needs the drift velocity to go from time (in pulses) to space (in the hits).
     */ 
    HitSplitter( IMPL::LCCollectionVec * SplitHitsCollection,
		 IMPL::LCCollectionVec * PulsesInSplitHitsCollection,
		 float vDrift);

    /** The split hits function returns a vector with the split hits. If the hit is not split the input
     *  hit is returned.
     *  
     *  Attention: The split hits that are produced are automatically written to the collection given
     *  to the constructor. If you store the hits in an additional collection make sure it has the 
     *  subSet flag turned on!
     *
     *  The idea behind this is that the hits which have not been split (which are usually already in 
     *  a collection) and the split ones can be written to one collection (the one with the subSet flag
     *  turned on). 
     *  
     */
    EVENT::TrackerHitVec splitHit( EVENT::TrackerHit const * const );
  
    protected:

    IMPL::LCCollectionVec * _splitHitsCollection;
    IMPL::LCCollectionVec * _pulsesInSplitHitsCollection;
    float _vDrift; ///< The drift velocity    

    /** A helper class to sort pulses according to their pad index.
     *  The sorting incudes the fact that there is a wraparround at phi=0 for circular geometry
     *  so i < j if abs(j - i) > nPadsInRow/2 and i > nPadsInRow/2 and j < nPadsInRow/2
     */

    class SortablePulse
    {
    public:
      bool operator<(const SortablePulse & p) const;
      int operator-(const SortablePulse & p) const;

      SortablePulse( EVENT::TrackerPulse * pulse);
      EVENT::TrackerPulse * getTrackerPulse() const;
    protected:
      EVENT::TrackerPulse * _pulse;
    };

    typedef std::pair< std::multiset<SortablePulse>::iterator, 
                       std::multiset<SortablePulse>::iterator> IteratorPair;

    /** A nested helper struct to contain two pulses and the distance between them, and 
     *  sort according to the distance.
     */
    struct Distance
    {
      float distance;
      std::multiset<SortablePulse>::iterator firstPulse;
      std::multiset<SortablePulse>::iterator secondPulse;
      
      Distance( float d, std::multiset<SortablePulse>::iterator first,
		std::multiset<SortablePulse>::iterator  second);
      
      bool operator<(Distance const &right) const { return distance < right.distance; }
    };
  
  };
}// namespace marlintpc

#endif // MARLINTPC_HIT_SPLITTER_H
