#ifndef PULSEFINDERPROCESSOR_H
#define PULSEFINDERPROCESSOR_H

//LCIO
#include "lcio.h"
#include "EVENT/LCCollection.h"
#include "EVENT/TrackerData.h"
#include <IMPL/TrackerDataImpl.h>
#include "EVENT/TrackerPulse.h"

//Marlin
#include "marlin/Processor.h"

//LCCD
#include "lccd/ConditionsMap.hh"
#include "Pedestal.h"
#include "GenericADCElectronicsParameters.h"

// MarlinTPC
#include "PedestalListener.h"
#include "ADCElectronicsListener.h"

//C++
#include <string>

namespace marlintpc
{

/** This processor creates lcio::TrackerPulse objects from lcio::TrackerData
 *
 *  First the pulses are searched using a threshold based method, where the start and end thresholds are parameters.
 *  Then these candidates are further processed, by splitting multiple pulses.
 *  In a last step quality flags are assigned to account for
 *    - multiple pulses (changed shape/charge information)
 *    - overflows, where the ADC values are saturated
 *    - anomalous shape, currently only too long pulses (FIXME: now 3 times the minimal pulse length)
 *  For the storage the integrated charge in ADC values is computed and the pulse time;
 *  the pulse time can be computed in two ways (choose by a parameter):
 *   (default:) the point of change in curvature in the rising slope
 *        or:   the charge weighted average/center of gravity
 *  As a default also the initial TrackerData (=ADC) spectrum is stored (choose by parameter),
 *  which is always forced when any quality flag is set.
 *
 *  <h4>Input - Prerequisites</h4>
 *  Collection of TrackerData (can be zero supressed)
 *
 *  <h4>Output</h4>
 *  Collection of TrackerPulses
 *  Collection of TrackerData which belong to a pulse
 *
 *  @param InputCollectionName:string name of the input TrackerData collection
 *  @param OutputTrackerPulseCollectionName:string name of the output TrackerPulse collection
 *  @param OutputTrackerDataCollectionName:string  name of the output TrackerData collection
 *  @param WriteOutputToStorage:bool output collection will be written if set to true (which is the default)
 *  @param ForceSpectrumSave:bool if not 0 the ADC spectrum will be stored for all pulses
 *  @param PulseStartThreshold:float a pulse start if ADC value is the given value times width of the pedestal
 *  @param PulseEndThreshold:float the pulse ends if the ADC value drops below this value
 *  @param MinimumPulseHeight:float to accept a pulse its maximal ADC entry must exceed this cut
 *  @param MinumumPulseLength:int to accept a pulse its length in time bins must be larger than this value without Start/End bins
 *  @param SaveNBinsBeforeStart:int # bins saved before the start bin (exept there is another pulse)
 *  @param SaveNBinsAfterStart:int # bins saved after the start bin (exept there is another pulse)
 *  @param ElectronicsParameterObjectName:string Name of the conditions object in which the parameters of the read out electronics is stored
 *  @param ElectronicsParametersCollectionName:string Name of the conditions collection in which the parameters of the read out electronics is stored
 *  @param UseParabolicFitAmplitudeAsCharge?:bool \c OPTIONAL Set this flag to use the amplitude calculated by fitting a parabola to the log of the pulse data
 *  @param UseParabolicFitMaxAsTime?:bool \c OPTIONAL Set this flag to use the time caclulated by fitting a parabola to the log of the pulse data
 *  @param MaximumADCValueOverride \c OPTIONAL: Manually set value of the maximum adc value
 *  @param PedestalWidthOverride \c OPTIONAL: Manually set value of the pedestal width
 *  @param SignalPolarityOverride \c OPTIONAL: Manually set value of the electronics polarity
 *  @param ReadoutFrequencyOverride \c OPTIONAL: Manually set value of the readout frequency
 *
 * @author Ch. Rosemann, DESY
 */
class PulseFinderProcessor : public marlin::Processor
{

    public:

        virtual Processor*  newProcessor()
        {
            return new PulseFinderProcessor;
        }

        PulseFinderProcessor();

        virtual void init();

        virtual void processRunHeader(lcio::LCRunHeader* run);

        virtual void processEvent(lcio::LCEvent* evt);

        virtual void check(lcio::LCEvent* evt);

        virtual void end();

    protected:

        int _nRun;          //< run counter
        int _nEvt;          //< event counter

        ///< name of the collection, where input data are stored
        std::string  _inputColName;
        ///< name of the collection, where TrackerData are stored
        std::string  _outputDataColName;
        ///< name of the collection, where TrackerPulse are stored
        std::string  _outputPulseColName;
        ///<- the name of the pedestal collection
        std::string  _pedestalColName;

        /// Name of the conditions object in which the parameters of the read out electronics is stored
        std::string  _electrParaObjName;

        /// Name of the conditions collection in which the parameters of the read out electronics is stored
        std::string _electrParaColName;

        /// Set this flag to use the amplitude (maximum bin value) of the pulse as the charge
        bool _use_amplitude_as_charge;

        /// set the persistency flag of the output collection
        bool _outputIsPersistent;

        /// if not 0 the ADC spectrum will be saved for all pulses
        bool _forceSpectrumSave;

        /** set if you want the default pulse time calculation (by using the sign change of the derivative of the charge spectrum)
        or if set 'true' then the time is calculated by fitting a parabola to the log of the bin values
        */
        bool _use_fit_time;
        bool _use_fit_charge;

        ///< start threshold to detect a pulse in adc values
        float _pulseStartThreshold;
        ///< end threshold to detect a pulse in adc values
        float _pulseEndThreshold;
        ///< minimal value for the maximum adc value
        float _minimumPulseHeight;
        ///< minimal length for a pulse in number of time bins
        int _minimumPulseLength;
        ///< # of bins stored before the start threshold
        int _saveNBinsBeforeStart;
        ///< # of bins stored after the end threshold
        int _saveNBinsAfterEnd;


        // override variables

        /** maximum adc value, needed to determine if a pulse is maxed out
         */
        float _maximumADCValueOverride;

        /** The readout frequency of every readout group (in Hz). This overrides any information provided by LCCD
         */
        float _readout_frequency_override;

        /** The pedestal width to be used for every hardware channel in every readout group. This overrides any information provided by LCCD
         */
        float _pedestal_width_override;

        /** the polarity of all readout groups, overrides LCCD
           */
        int _polarity_override;

        // the default way of providing the access to the pedestal values:
        PedestalListener* _pedestalListener;
        // the default way to access the electronics parameters
        ADCElectronicsListener* _adcElectronicsListener;


        /** some helper classes to facilitate the calculation of the pulse properties: */
        class _pulseTime
        {
            public:
                _pulseTime() : _time(0.), _timeError(0.)
                {
                    ;
                };
                void setTime(float t)
                {
                    _time = t;
                };
                void setTimeError(float e)
                {
                    _timeError = e;
                };
                float getTime()
                {
                    return _time;
                };
                float getTimeError()
                {
                    return _timeError;
                };
            private:
                float _time, _timeError;
        };

        class _pulseCharge
        {
            public:
                _pulseCharge() : _charge(0.), _chargeError(0.)
                {
                    ;
                };
                void setCharge(float t)
                {
                    _charge = t;
                };
                void setChargeError(float e)
                {
                    _chargeError = e;
                };
                float getCharge()
                {
                    return _charge;
                };
                float getChargeError()
                {
                    return _chargeError;
                };
            private:
                float _charge, _chargeError;
        };

        class _pulse : public _pulseCharge, public _pulseTime
        {
            public:
                _pulse() : _pulseCharge(), _pulseTime()
                {
                    ;
                };
                void add(_pulseCharge c)
                {
                    setCharge(c.getCharge());
                    setChargeError(c.getChargeError());
                };
                void add(_pulseTime t)
                {
                    setTime(t.getTime());
                    setTimeError(t.getTimeError());
                };
        };

        /** another helper class to make more sense to the splitting phase:
         *  actually keep track of the "split" flag
          */
        struct internalSplitPulse
        {
            internalSplitPulse(std::vector<float>::iterator s, std::vector<float>::iterator e, bool f)
            {
                startBin = s;
                endBin = e;
                isSplit = f;
            }
            std::vector<float>::iterator startBin;
            std::vector<float>::iterator endBin;
            bool isSplit;
private:
            internalSplitPulse()
            {
                ;
            };
        };

        /** Find pulses by threshold in the ADC information and closer inspection.
        *  This is the somewhat "master" function that determines all aspects of the pulses,
        *  it calls some further functions for easier reading.
        *  @param thisData 'raw' data as TrackerData
        *  @param tempPulse a helper data construct, a pair of the TrackerData of a pulse and the quality bit
        *  @param noiseValue the width of the pedestal
         */
        void findPulses(lcio::TrackerData* thisData, std::vector<std::pair<lcio::TrackerDataImpl*, int> >* tempPulse, float noiseValue);



        /** The first function that is called during the pulse finding,
        it simply looks for values above the start threshold and below the end thresholds,
        the start and end positions are stored by their iterators in the original TrackerData ADC value vector.
        Here already pulses which are too short are suppressed.
        @param ADCSpectrum the ADC spectrum to be searched
        @param theCandidatePulses a vector of the pairs of start and end positions of the pulse candidates (as iterators)
        */

        void brutePulseSearch(std::vector<float>& ADCSpectrum,
                              std::vector<std::pair<std::vector<float>::iterator, std::vector<float>::iterator> >* theCandidatePulses, const int polarity);



        /** The second part in the pulse finding process.
        Check the pulse candidates for possible multiple pulses by looking more closely at the ADC spectrum.
        Then replace the /old/ pairs of start/end iterators with the appropriate new number of pairs.
        TODO: adjust the charges in the bins at the joined boundary of the split pulses
        @param theCandidatePulses a vector of the pairs of start and end positions of the pulse candidates (as iterators)
        @param noiseValue the pedestal width
        */

        std::vector<internalSplitPulse> splitMultiplePulses(std::vector<std::pair<std::vector<float>::iterator, std::vector<float>::iterator> >*
                                                            theCandidatePulses, float noiseValue);



        /** This wrapper function handles the calculation of the pulse time and the pulse charge.
         *  They are combined into one function because it is faster to calculate both values at the same time if the fit method is used.
         *  It can also be modified to add new methods of calculating these values without modifying the main event loop.
         *  New addition: errors on all the values!
         *  Convention: first time and time error, then charge and charge error
        @param data the pulse data
        @param noise the noise value
        */
        _pulse calculatePulseTimeAndCharge(const EVENT::TrackerData *tracker_data);



        /** This function calculates the pulse time and charge by fitting a parabola to the log of the data surrounding the extreme point.
        @param data the pulse data
        */
        _pulse calculatePulseWithFit(const EVENT::TrackerData *tracker_data);



        /** Calculate the time information of the given pulse spectrum.
         *  Two different ways are implemented, which can be chosen by a processor parameter.
         *  The default way is the point of change in the curvature of the rising slope of the pulse.
         *  The second possibility is to take the charge weighted mean of the time (center of gravity).
         *  @param thisPulse ADC spectrum from which the time will be calculated
         *  @param NoiseValue variation smaller than this value will be ingnored during the calculation
         *  @return the time information of the pulse in ns
         */
        _pulseTime calcPulseTime(const lcio::TrackerData* thisPulse);



        /** Calculate the integrated charge of the given pulse by summing up all bins of the ADC spectrum.
         *  The polarity of the pulse is taken into acount
         *  @param thisPulse ADC spectrum for charge calculation
         */
        _pulseCharge calcPulseCharge(const lcio::TrackerData* thisPulse);



        /** Returns the pedestal width for a given TrackerData / TrackerPulse object taking into consideration whether it should be overwritten by a processor parameter
         */
        float getPedestalWidth(const int hardware_channel, const int readout_group);



        /** Returns the polarity of a TrackerData / TrackerPulse object taking into consideration whether is should be overwritten by a processor parameter
         */
        int getPolarity(const int readout_group);



        /** Returns the readout frequency of a readout group (in Mhz) taking into consideration whether it should be overwritten by a processor parameter
         */
        float getReadoutFrequency(const int readout_group);



        /** Returns the maximum ADC value of a readout group taking into consideration whether it should be overwritten by a processor parameter
         */
        float getMaxADCValue(const int readout_group);


};
}
#endif // PULSEFINDERPROCESSOR
