#ifndef TRACK_FITTER_SIMPLE_CHI_SQUARE_PROCESSOR_H
#define TRACK_FITTER_SIMPLE_CHI_SQUARE_PROCESSOR 1

#include "marlin/Processor.h"
//#include "lcio.h"
//#include <string>
#include <vector>
#include "TrackFitterFactory.h"

//#include <EVENT/TrackerHit.h>

//namespace AIDA {
//  
//  class IHistogram1D;
//  class IHistogram2D;
//
//
//}


namespace marlintpc
{

/** A processor which instantiates a TrackFitterSimpleChiSquare and fits the tracks with it.
 *  @author M. Killenberg(University of Bonn)
 *
 *  <h4>Input - Prerequisites</h4>
 *  \li A collection of track candidates (lcio::Tracks) containing the TrackerHits.<br>
 *	Note: The LCIO::TRBIT_HITS flag has to be set when this collection is created so
 *      the hits are stored in the track.
 *
 *  <h4>Output</h4>
 *  \li Collection of fitted tracks.
 *
 *  @param InputTrackCandidates The the name of the input collection of track candidates (default: TPCTrackCandidates)
 *  @param OutputTracks The name of the collection the straight tracks be stored under (default: TPCTracks)
 *  @param SetOutputTransient If not 0 the output collection is set transient (default: 0)
 *  @param TransDefocussing The transverse defocussing in the readout structure \f$\sigma_{0,{\mathrm trans}}\f$ (in mm) (default: 1)
 *  @param LongDefocussing  The longitudinal defocussing in the readout structure \f$\sigma_{0,{\mathrm long}}\f$ (in mm) (default: 1)
 *  @param TransDiffusionCoef The transverse diffusion coefficient (in mm/\f$\sqrt{\mathrm mm}\f$) (default: 0)
 *  @param TransDefocussing  The longitudinal diffusion coefficient (in mm/\f$\sqrt{\mathrm mm}\f$) (default: 0)
 *
 *  The parameters for the starting values of the fit are optional parameter. If the parameter is given, it is use.
 *  If the parameter is not given the value from the seed track is used.
 *  
 *  @param D0Start \c optional Use this start value for D0 instead of the value from the seed track
 *  @param TanLambdaStart \c optional Use this start value for TanLambda instead of the value from the seed track
 *  @param Z0Start \c optional Use this start value for Z0 instead of the value from the seed track
 *  @param OmegaStart \c optional Use this start value for Omega instead of the value from the seed track
 *  @param PhiStart \c optional Use this start value for Phi instead of the value from the seed track
 *  
 *  The following parameters allow to fix a track parameter during the fit. To fit a straight line for instance set
 *  \c OmegaStart to 0. and \cc FixOmega to \c true.
 *
 *  @param FixD0        Do not vary d0        when fitting (default: false)
 *  @param FixPhi       Do not vary phi       when fitting (default: false)
 *  @param FixOmega     Do not vary omega     when fitting (default: false)
 *  @param FixTanLambda Do not vary tanLambda when fitting (default: false)
 *  @param FixZ0        Do not vary z0        when fitting (default: false)
 *
 *  If you just want to fit in the transverse direction set NoZFit to true. The residuals function will not include the residuals in z.
 *
 *  @param NoZFit       Do not fit in z direction, only use residuals in x-y  / r-phi (default: false)
 *
 */
class TrackFitterSimpleChiSquareProcessor : public marlin::Processor {
  
public:

	/** Default constructor
	*/
	TrackFitterSimpleChiSquareProcessor();


	/** Destructor
	*/
	~TrackFitterSimpleChiSquareProcessor();


	/** Return a new instance of this processor
	*/
	Processor * newProcessor();


	/** In this processor the init function set the 
	 *  collection parameters for the track collection 
	 *  and instantiates the TrackFitterFactory
	*/
	void init();


	/** Called once per run to process the event's header
	*/
	void processRunHeader (EVENT::LCRunHeader *run);


	/** Process event requests a track fitter from the 
	 *  TrackFitterFactory for each event (to update possible 
	 *  conditions data).
	 *  Afterwards it loops all tracks and fits them.
	 */
	virtual void processEvent(EVENT::LCEvent *evt);


	/** Called for every event - right after processEvent() 
	 *has been called for all processors.
	*/
	virtual void check (EVENT::LCEvent *evt);


	/** Called after data processing for clean up in the 
	 *  inverse order of the init() method so that resources 
	 *  allocated in the first processor also will be available for 
	 *  all following processors.
	*/
	virtual void end();


protected:

	/** The name of the input collection of seed tracks
	*/
	std::string _input_seed_tracks_collection_name;

	/** The name of the output collection with the fitted tracks
	*/
	std::string _output_tracks_collection_name;

	/** Give the status which is set to the transient flag of the output collection.
	 *  It is an int instead of a boolean, because the processor parameter 
	 *  cannot be a boolean.
	 *  Hits and tracks can be set separately.
	 */
	int _outputIsTransient; //< if not 0 the hits collection is set transient 
	                        //(default 0)

	EVENT::LCParameters * _trackCollectionParameters;

	float _transDefocussing, _longDefocussing, _transDiffusionCoef, _longDiffusionCoef;

	/// the optional start parameters
	float _d0Start, _phiStart, _omegaStart, _tanLambdaStart, _z0Start;

	/// the flags which fit parameters to fix
	bool _fixD0, _fixPhi, _fixOmega, _fixTanLambda, _fixZ0;

	/// flag whether to fit in z or only xy
	bool _noZFit;

	int _nGoodFits, _nBadFits;
};

}// namespace marlintoc

#endif // TRACKSEEDERPROCESSOR_H



