#ifdef USE_LCCD

// MarlinTPC
#include "GainCorrectorProcessor.h"
#include "ChannelCorrectionHandler.h"

// C++
#include <iostream>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerPulse.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerPulseImpl.h>
#include <IMPL/LCFlagImpl.h>

// tpcconddata
#include "ChannelCorrection.h"

using namespace lcio ;
using namespace marlin ;
using namespace std;
using namespace tpcconddata;

namespace marlintpc {


GainCorrectorProcessor aGainCorrectorProcessor;


GainCorrectorProcessor::GainCorrectorProcessor() : Processor("GainCorrectorProcessor"),

	_input_tracker_pulses_collection_name("TPCPulses"),
	_output_tracker_pulses_collection_name("TPCPulses"),
	_channel_corrections_collection_name( tpcconddata::ChannelCorrection::getDefaultColName() ), 
	_output_is_transient(true),
	_channel_corrections(NULL) {

	/* description of the processor which will be displayed in the
	* steering file autometicly generated by
	* marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
	*/

	_description = "GainCorrectorProcessor converts FADC counts into electron counts, assumes a linear gain (needing no corrections), provided via LCCD.";

	/* register steering parameters: name, description, class-variable, default value
	* the type will by definded by the type of the default value
	* string, double, float and int are possible
	* plaese note to get a float you must cast in C++: (float) 1.3
	*/
	registerInputCollection( LCIO::TRACKERDATA ,
					"InputTrackerPulses",
					"Name of the input collection as stored in the LCIO file",
					_input_tracker_pulses_collection_name,
					std::string("TPCPulses") );

	registerOutputCollection( LCIO::TRACKERDATA ,
					"OutputTrackerPulses",
					"Name of the collection to be stored in the LCIO file",
					_output_tracker_pulses_collection_name,
					std::string("TPCPulses") );

	registerOptionalParameter( "SetOutputTransient",
				    "if not 0 the output collection is set transient",
				    _output_is_transient,
				    true);

	registerOptionalParameter( "ChannelCorrectionCollection" ,
				    "Name of the collection containing the channel corrections",
				    _channel_corrections_collection_name,
				    tpcconddata::ChannelCorrection::getDefaultColName() );
}


void GainCorrectorProcessor::init() {
	
	// usually a good idea to
	printParameters();
	
	_number_of_runs = 0;
	_number_of_events = 0;

	_channel_corrections = new ChannelCorrectionHandler( _channel_corrections_collection_name );
}
    

void GainCorrectorProcessor::processRunHeader( LCRunHeader* run) {
	
	_number_of_runs++ ;
}

void GainCorrectorProcessor::processEvent( LCEvent * evt ) {
	
	// get the input collection from the event
	try {

		LCCollection* input_pulses = evt->getCollection( _input_tracker_pulses_collection_name );

		streamlog_out( MESSAGE3 ) << "Processing " << input_pulses->getNumberOfElements() << " pulses.\n.";

	    // create the output collection with a string defining the TYPE (see lcio.h)
	    LCCollectionVec* output_pulses = new LCCollectionVec( LCIO::TRACKERPULSE );
	    // set the transient flag as definded by the processor parameter
	    output_pulses->setTransient( (_output_is_transient != 0) );
		output_pulses->setSubset( false );

	    // set the flags that cellID1 should be stored
	    lcio::LCFlagImpl trkFlag(0) ;
	    trkFlag.setBit( lcio::LCIO::TRAWBIT_ID1 ) ;
	    output_pulses->setFlag( trkFlag.getFlag() );

	    for(int iRawData = 0; iRawData< input_pulses->getNumberOfElements(); iRawData++) {

			/* getElement return only a LCObject
			* you have to cast it the get the right type
			* please use dynamic_cast which will return 0 if the object can not be casted
			*/
			TrackerPulse* input_pulse = dynamic_cast<TrackerPulse*>( input_pulses->getElementAt( iRawData ) ) ;

			if ( !input_pulse ) {

				throw( Exception("Input collection " + _input_tracker_pulses_collection_name + " does not contain TrackerPulses.") );
			}


			// the first channel correction parameter stores the gain of the signal shaping amplifier which converted electron counts into voltage levels
			const double total_gain = _channel_corrections->getChannelCalibrationFactor( input_pulse->getCellID0(), 0 );

			/* create a new object of the TrackerData type
			* you have to use the IMPL, otherwise you can not set values
			*/
			TrackerPulseImpl* output_pulse = new TrackerPulseImpl();

			// transver the values

			//first the simple
			output_pulse->setCellID0( input_pulse->getCellID0() );
			output_pulse->setCellID1( input_pulse->getCellID1() );// normally 0 for TPC data
			output_pulse->setTime( input_pulse->getTime() );
			output_pulse->setQuality( input_pulse->getQuality() );
			output_pulse->setTrackerData( input_pulse->getTrackerData() );

			// now convert from fadc values to primary electrons
			output_pulse->setCharge( input_pulse->getCharge() / total_gain );

			streamlog_out( DEBUG2 ) << "Pad #" << input_pulse->getCellID0() << ", amplified electrons: " << input_pulse->getCharge() << ", original electrons: " << output_pulse->getCharge() << ".\n";

			// add the converted pulse to the collection
			output_pulses->addElement( output_pulse );
	    }

	    
	    /* after convertin all raw data objects
	     * add the collection to the eventand asigning a name to it
	     * please note again: DO NOT DELETE the collection  this will be done by LCIO
	     */

		evt->addCollection( output_pulses, _output_tracker_pulses_collection_name );
	}
	catch( DataNotAvailableException e ) {
	    // an error occured
	    streamlog_out( ERROR ) << e.what() << std::endl;
	}

	_number_of_events++;
}
    
    

void GainCorrectorProcessor::check( LCEvent * evt ) {
	// nothing to check here - could be used to fill checkplots in reconstruction processor
}
    
    
void GainCorrectorProcessor::end() {

	if( _channel_corrections ) {

		delete _channel_corrections;
		_channel_corrections = NULL;
	}
}

}

#endif // USE_LCCD
