#include "HitCandidate.h"

#include <gear/TPCParameters.h>
#include <gear/TPCModule.h>
#include <EVENT/TrackerPulse.h>
#include <streamlog/streamlog.h>


using namespace gear;
using namespace lcio;

namespace marlintpc
{
  HitCandidate::HitCandidate(const TPCParameters & tpcParams, double vDrift, 
			     TrackerPulse* pulse)
    : _maxPulse(0), _maxPulseCharge(0.0), _numberOfDeadPads(0), _isAtBoundary(false) ,
      _vDrift( vDrift ), 
      _tpcParameters( tpcParams ), _tpcModule(0) ,
      _localCoordinates(3), _globalCoordinates(3) ,
      _globalCovariance(6), _charge(0), _localFixedVariance(0), _localFreeVariance(0),
      _localSkewness(0),
      _zVariance(0), _hitParametersAreValid(false)
  {
    if (pulse) addPulse( pulse );
    
    _localCoordinates.assign(   _localCoordinates.size(), 0 );
    _globalCoordinates.assign( _globalCoordinates.size(), 0 );

    // set the covariance matrix to unity (lower triangle)
    _globalCovariance[0]=1; //xx or rr
    _globalCovariance[1]=0; //xy or rphi
    _globalCovariance[2]=1; //yy or phiphi
    _globalCovariance[3]=0; //xz or rz
    _globalCovariance[4]=0; //yz or phiz
    _globalCovariance[5]=1; //zz 

  }
    
  void HitCandidate::addPulse(TrackerPulse* aPulse)
  {
    _thePulses.push_back(aPulse);
    if(fabs(aPulse->getCharge()) > fabs(_maxPulseCharge))
    {
      _maxPulseCharge = aPulse->getCharge();
      _maxPulse       = aPulse;
      // get the TPCModule for this pulse
      _tpcModule = & (_tpcParameters.getModule( aPulse->getCellID1() ));
    }

    _hitParametersAreValid=false;
  }

  void HitCandidate::addDeadPad()
  {
    _numberOfDeadPads++;
  }
  
  void HitCandidate::setBoundaryFlag()
  {
    _isAtBoundary = true;    
  }
    
  double HitCandidate::getCharge()
  {
    // to calculate all parameters seems overdone just for the charge
    // but the calculation (loop) is only run once for all parameters
    if ( !_hitParametersAreValid) calculateHitProperties();

    return _charge;
  }

  double HitCandidate::getMaxPulseHeight() const
  {
    return _maxPulseCharge;
  }

  unsigned int HitCandidate::getWidth() const
  {
    return _thePulses.size();
  }

  double HitCandidate::getLocalFixedCoordinate()
  {
    if ( !_hitParametersAreValid) calculateHitProperties();

    // return y or r, depending on coordinate system
    if(_tpcModule->getLocalPadLayout().getCoordinateType() == PadRowLayout2D::CARTESIAN)
      return _localCoordinates[1]; // this is truly y
    else if(_tpcModule->getLocalPadLayout().getCoordinateType() == PadRowLayout2D::POLAR)
      return _localCoordinates[0]; // this is R
    else
      throw gear::Exception("Coordinate type unknown!");
  }

  double HitCandidate::getLocalFreeCoordinate()
  {
    if ( !_hitParametersAreValid) calculateHitProperties();

    // return x or phi, depending on coordinate system
    if(_tpcModule->getLocalPadLayout().getCoordinateType() == PadRowLayout2D::CARTESIAN)
      return _localCoordinates[0]; // this is truly x
    else if(_tpcModule->getLocalPadLayout().getCoordinateType() == PadRowLayout2D::POLAR)
      return _localCoordinates[1]; // this is phi
    else
      throw gear::Exception("Coordinate type unknown!");
  }
 
  double HitCandidate::getZCoordinate()
  {
    if ( !_hitParametersAreValid) calculateHitProperties();

    return _localCoordinates[2];
  }

  double HitCandidate::getLocalFixedVariance()
  {
    if ( !_hitParametersAreValid) calculateHitProperties();

    return _localFixedVariance;    
  }

  double HitCandidate::getLocalFreeVariance()
  {
    if ( !_hitParametersAreValid) calculateHitProperties();

    return _localFreeVariance;    
  }

  double HitCandidate::getZVariance()
  {
    if ( !_hitParametersAreValid) calculateHitProperties();

    return _zVariance;    
  }

  lcio::DoubleVec HitCandidate::getLocalCoordinates()
  {
    if ( !_hitParametersAreValid) calculateHitProperties();

    return _localCoordinates;    
  }

  lcio::DoubleVec HitCandidate::getGlobalCoordinates()
  {
    if ( !_hitParametersAreValid) calculateHitProperties();

    return _globalCoordinates;    
  }

  lcio::FloatVec HitCandidate::getGlobalCovarianceMatrix()
  {
    if ( !_hitParametersAreValid) calculateHitProperties();

    return _globalCovariance;    
  }

  void HitCandidate::calculateHitProperties()
  {
    // don't do anyting if there are no pulses 
    // (especially avoid divide by zero in the calculations by doing nothing ;-) )
    if (_thePulses.size() == 0) return;

    // get the local pad layout from the module.
    const gear::PadRowLayout2D &localPadPlane = _tpcModule->getLocalPadLayout();
    
    double zAnode=0;
    // Set the correct position of the anode from the gear file.
    // This is placed here and not class wide because it will be module dependent in the future.
    try
    {
      zAnode =  _tpcParameters.getDoubleVal ("zAnode");
    }
    catch(gear::UnknownParameterException &)
    {
      // no Anode position set in the GEAR file, use max drift length instead
      streamlog_out(WARNING) << "No zAnode tag found in gear file. Using maxDriftLength instead."
			     << std::endl;
      zAnode = _tpcParameters.getMaxDriftLength ();
    }
    

    double sum_xphi_q =0.; // sum(x*q) or sum(phi*q)
    double sum_xphi_xphi_q =0.; // sum(x*x*q) or sum(phi*phi*q)
    double sum_xphi_xphi_xphi_q =0.; // sum(x*x*x*q) or sum(phi*phi*phi*q)
    double sum_z_q      =0.; // sum(z*q)
    double sum_z_z_q      =0.; // sum(z*z*q)
    double sumq       =0.; // sum(q)
    double xphi       =0.; // coordinate along pad row (free coordinate)
    double yr         =0.; // coordinate orthogonal to pad row (fixed coordinate)

    
    // loop over all pulses in cluster and sum up
    for (std::vector<EVENT::TrackerPulse*>::iterator pulseIter = _thePulses.begin();
	 pulseIter < _thePulses.end(); pulseIter++)
    {
      // dereference the Iterator for code readability (iterator of pointer is a bit ugly)
      TrackerPulse* pulse=*pulseIter;

      // get pad coordinates of current pulse from GEAR layout
      Vector2D padCenterLocal = localPadPlane.getPadCenter( pulse->getCellID0() );

      // get x / phi coordinate 
      switch ( localPadPlane.getCoordinateType() )
      {
        case PadRowLayout2D::CARTESIAN:
	    xphi = padCenterLocal[0]; // x
	    break;
        case PadRowLayout2D::POLAR:
	    xphi = padCenterLocal[1]; // phi
	    break;
        default:
	  throw gear::Exception("TopoFinder: unkown coordinate type");
      };

      // take y / r coordinate from first pad, it's the same for all pads
      if ( pulseIter == _thePulses.begin() )
      {
	switch ( localPadPlane.getCoordinateType() ) 
	{
	  case PadRowLayout2D::CARTESIAN:
	    yr = padCenterLocal[1]; // y
	    break;
	  case PadRowLayout2D::POLAR:
	    yr = padCenterLocal[0]; // r
	    break;
	  default:
	    throw gear::Exception("TopoFinder: unkown coordinate type");
	}; // switch
      } // if first pulse

      // summ correctly if handling pads from the first and thr fourth quadrant 
      // in polar coordinates (wrap around from 2 pi to 0)
      if ( (pulseIter != _thePulses.begin()) // for the first pad there is nothing to compare with
	   && (localPadPlane.getCoordinateType() == PadRowLayout2D::POLAR) ) // problem only occurs in polar coordinates
       {
	 // there can be thwo cases
	 //1.
	 if ( (sum_xphi_q /sumq < M_PI_2) // pulse up to now is in first quadrant (or in fourth, already mapped to <0)
	       && (xphi > 1.5*M_PI) ) // current pulse in fourth quadrant
	   xphi-=2*M_PI; // map to range [-pi:pi] so averaging works

	 // 2.
	 // no else needed; statements are mutually excluse because due to first condition
	 // even if xphi was changed in the first statement
	 if ( (sum_xphi_q /sumq > 1.5* M_PI) // pulse up to now is in fourth quadrant (or in first, already mapped to >2*pi)
	       && (xphi < M_PI_2) ) // current pulse in first quadrant
	   xphi+=2*M_PI; // map to range [pi:3*pi] so averaging works
       }

      // add to the sums
      sum_xphi_q += xphi * pulse->getCharge();  // x * q or phi * q
      sum_xphi_xphi_q += xphi * xphi * pulse->getCharge();  // x * q or phi * q
      sum_xphi_xphi_xphi_q += xphi * xphi * xphi * pulse->getCharge();  // x * q or phi * q
      sum_z_q += pulse->getTime() * _vDrift / 1000 // t * vDrift = z // 1000 ns/us
	* pulse->getCharge() ;   // * q
      sum_z_z_q += pulse->getTime() * _vDrift / 1000 // t * vDrift = z // 1000 ns/us
	* pulse->getTime() * _vDrift / 1000 // t * vDrift = z // 1000 ns/us
	* pulse->getCharge() ;   // * q
      sumq +=pulse->getCharge() ;     // q
      
    }// for (pulse)

    // calculate the local coordinates and store them
    switch ( localPadPlane.getCoordinateType() ) 
    {
      case PadRowLayout2D::CARTESIAN:
	// in cartesian coordinates xphi is x, yr is y
	_localCoordinates[0] =  sum_xphi_q / sumq;
	_localCoordinates[1] =  yr;
	_localCoordinates[2] =  zAnode - sum_z_q / sumq;
	break;
      case PadRowLayout2D::POLAR:
	_localCoordinates[0] =  yr; // it is r
	_localCoordinates[1] =  sum_xphi_q / sumq; // it is phi
	_localCoordinates[2] =  zAnode - sum_z_q / sumq;
	break;
    };

    // let gear do the conversion to global coordinates for us
    gear::Vector2D positionGlobal2D = _tpcModule->localToGlobal(_localCoordinates[0],
							       _localCoordinates[1]);

    _globalCoordinates[0] = positionGlobal2D[0];
    _globalCoordinates[1] = positionGlobal2D[1];
    _globalCoordinates[2] = _localCoordinates[2];
    
    // calculate the local and global errors
    /// \FIXME Implement calcuclation of errors
 }

  TrackerHitImpl * HitCandidate::createTrackerHit()
  {
    if (_thePulses.size()==0)
    { // the hit candidate is empty, no point in creating a hit
      return 0;
    }

    if ( !_hitParametersAreValid) calculateHitProperties();
    
    TrackerHitImpl * newHit =  new TrackerHitImpl;
    
    double pos[3];
    
    // lcio coordinates are always cartesian, convert if necessary
    switch ( _tpcModule->getTPCCoordinateType() ) 
    {
      case PadRowLayout2D::CARTESIAN:
	// no conversion needed
	pos[0] = _globalCoordinates[0];
	pos[1] = _globalCoordinates[1];
	break;
      case PadRowLayout2D::POLAR:
	// convert polar to cartesian
	pos[0] = _globalCoordinates[0]*cos( _globalCoordinates[1] );
	pos[1] = _globalCoordinates[0]*sin( _globalCoordinates[1] );
	break;
      default:
	throw gear::Exception("TopoFinder: unkown coordinate type");
    }; // switch

    pos[2] = _globalCoordinates[2];

    newHit->setPosition (pos);
    newHit->setCovMatrix ( _globalCovariance );
    newHit->setdEdx ( _charge );
    
    // if (_numberOfDeadPads) set contains empty pads flag
    // if one of the pulses has the split flag set split flag
    // if one of the pulses has been interpolated set interpolated flag

    return newHit;
  }

}// namespace marlintpc
