#include "HitSplitter.h"
#include "HitCandidate.h"
#include <iostream>
#include <string>
#include <algorithm>
#include <set>
#include <cmath>

// include what you need from lcio
#include <EVENT/LCEvent.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerHitImpl.h>
#include <IMPL/TrackerPulseImpl.h>
#include <EVENT/TrackerData.h>
#include <Exceptions.h>

#include <marlin/Global.h>
#include <streamlog/streamlog.h>

#include <gear/TPCModule.h>

#include "SignalSplitter.h"

using namespace lcio;
using namespace marlin;
using namespace std;


namespace marlintpc
{

  HitSplitter::HitSplitter( LCCollectionVec * splitHitsCollection,
			    LCCollectionVec * pulsesInSplitHitsCollection,
			    float vDrift)
    : _splitHitsCollection(splitHitsCollection),
      _pulsesInSplitHitsCollection(pulsesInSplitHitsCollection ),
      _vDrift(vDrift)
  {
  }

  TrackerHitVec HitSplitter::splitHit( TrackerHit const * const inputHit )
  { 
    TrackerHitVec outputVec;

    // fill all the pulses in a map, sorted by channel.
    std::multiset<SortablePulse> pulseSet;
    for (LCObjectVec::const_iterator pulseIter = inputHit->getRawHits().begin();
	 pulseIter <  inputHit->getRawHits().end(); pulseIter++)
    {
      SortablePulse pulse( dynamic_cast<TrackerPulse *>(*pulseIter) ); 

      pulseSet.insert(pulse);
    }

    // Check if there are two pulses on at least one of the channels
    // If this is the case perform regrouping of pulses
    bool regroup=false;
    { 
      std::multiset< SortablePulse >::iterator firstIter= pulseSet.begin();
      std::multiset< SortablePulse >::iterator secondIter= firstIter;
      ++secondIter;
      
     while ( secondIter != pulseSet.end() )
      {
	if ( firstIter->getTrackerPulse()->getCellID0() == secondIter->getTrackerPulse()->getCellID0() )
	{
	  regroup=true;
	  break;
	}
	++firstIter;
	++secondIter;
      }
    }

    std::vector< std::vector< SortablePulse > > preCandidates;    
    if ( regroup )
    {
      streamlog_out(DEBUG3) << "This is regroup"<< std::endl;
      // fill the distances between pulses in neighbouring channels into vectors
      IteratorPair firstRange = pulseSet.equal_range( *(pulseSet.begin()) );
      std::vector< std::set< Distance > > distancesInChannels;
            
      while ( firstRange.second != pulseSet.end() )
      {
	// calculate all the distances to all the neighbouring pulses
	// this is safe since we checked that firstRange.second is a valid pulse and not end()
	IteratorPair secondRange = pulseSet.equal_range( *(firstRange.second));

	streamlog_out(DEBUG3) << "Looping both ranges..."<< std::endl;
      
	std::set< Distance > distances;
	for ( std::multiset< SortablePulse >::iterator firstIter = firstRange.first ;
	      firstIter != firstRange.second ; ++ firstIter )
	{
	  for ( std::multiset< SortablePulse >::iterator secondIter = secondRange.first ;
		secondIter != secondRange.second ; ++ secondIter )
	  {
	    // check that pulses really overlap (to avoid association 1 to 4)
	    //
	    //  1
	    //  1
	    //  1 3
	    //  2 3 
	    //  2 3
            //  2 4
            //    4
	    float start1 = firstIter->getTrackerPulse()->getTrackerData()->getTime();
	    float end1 = start1 + firstIter->getTrackerPulse()->getTrackerData()->getChargeValues().size();
	    float start2 = secondIter->getTrackerPulse()->getTrackerData()->getTime();
	    float end2 = start2 + secondIter->getTrackerPulse()->getTrackerData()->getChargeValues().size();
	    
	    if ( !( ( end2< start1) ||  (start2 > end1) ) )
	    {
	      // pulses overlap, fill the distance
	      distances.insert( Distance(std::fabs( firstIter->getTrackerPulse()->getTime()
						    - secondIter->getTrackerPulse()->getTime() ),
					 firstIter,
					 secondIter ) );
	    }
	  }// for seondIter
	}// for firstIter

	streamlog_out(DEBUG3) << "done. Found " <<distances.size() << " distances"<< std::endl;

	streamlog_out(DEBUG3) << "Looping the distances... " << std::endl;

	// now only keep the shortest distances
	std::set< Distance >::iterator distIter= distances.begin();
	while( distIter != distances.end() )
	{
	  std::set< Distance >::iterator nextDistIter = distIter;
	  ++nextDistIter;
	  if ( nextDistIter == distances.end() ) break;

	  // throw out all the distance object wich contains one of the two pulses in distIter
	  // (as distIter is the combination with the shortest distance the others can only be larger)
	  while( nextDistIter != distances.end() )
	  {
	    if ( ( nextDistIter->firstPulse == distIter->firstPulse ) ||
		 ( nextDistIter->secondPulse == distIter->secondPulse ) )
	    {
	      // it is safe to delete an iterator like this. The ++ is evaluated before the erase,
	      // so it operates on a valid iterator
	      distances.erase(nextDistIter++);
	    }
	    else
	    {
	      // increase the iterator without erasing
	      ++nextDistIter;
	    }
	      
	  } // while nextDistIter

	  // after deleting the invalid distance objects we can increase the distIter
	  ++distIter;
	}
	
	streamlog_out(DEBUG3) << "done. " <<distances.size() 
			      << " distances left after keeping only the shortest"<< std::endl;

	// now only the iterator pairs we want are in the distances set.
	// This channel is complete. Add the distaces set to the distancesInChannels vector
	// and continue with next channel
	distancesInChannels.push_back( distances );
	firstRange = secondRange;

      }//while ( firstRange.second != pulseSet.end() )

      streamlog_out(DEBUG3) << "Found " <<distancesInChannels.size() 
			    << " distances in total."<< std::endl
			    << " filling the hit candidates "<< std::endl;
      

      // fill the hit candidates by searching all pulse pairs
      for (std::vector< std::set< Distance > >::iterator channelIter = distancesInChannels.begin();
	   channelIter < distancesInChannels.end(); channelIter++)
      {
	for (std::set<Distance>::iterator distIter = channelIter->begin(); 
	     distIter != channelIter->end(); ++distIter)
	{
	  // create a new hit candidate
	  std::vector< SortablePulse >  preCandidate;

	  //fill the first two pulses into the candidate
	  preCandidate.push_back( *(distIter->firstPulse) );
	  preCandidate.push_back( *(distIter->secondPulse) );

	  std::multiset<SortablePulse>::iterator lastAddedPulse = distIter->secondPulse;
	  
	  // search in the next channel for a matching pulse to the last added one
	  for ( std::vector< std::set< Distance > >::iterator channelToBeSearched = channelIter+1;
		channelToBeSearched < distancesInChannels.end() ; 
		/*don't increase here because of erase in loop*/ )

	  {
	    bool pulseWasFound = false;
	    for ( std::set< Distance >::iterator distanceToBeCompared = channelToBeSearched->begin();
		  distanceToBeCompared != channelToBeSearched->end(); ++distanceToBeCompared)
	    {
	      if ( distanceToBeCompared->firstPulse ==  lastAddedPulse )
	      {
		preCandidate.push_back( *(distanceToBeCompared->secondPulse) );
		lastAddedPulse=distanceToBeCompared->secondPulse;
		// remove the distanceToBeCompared from the list of available pulses
		// this is safe since the loop is quit with break afterwards
		channelToBeSearched->erase( distanceToBeCompared );

		pulseWasFound=true;
		break;
	      }
	    }

	    if (pulseWasFound)
	    {
		if ( channelToBeSearched->empty() )
		{
		  // this sets the channelToBeSearched to the next valid iterator
		  channelToBeSearched=  distancesInChannels.erase(channelToBeSearched);
		}
		else // take the next iterator without deleting the current one
		{
		  channelToBeSearched++; 
		}
	    }
	    else{
	      break;}

	  }// for channelToBeSearched

	  // add the hit candidate to the vector of hit candidates
	  preCandidates.push_back( preCandidate );
	}//for distIter
      }//for channelIter

      streamlog_out(DEBUG3) << "End of regroup, found " 
			    << preCandidates.size() << " preCandidates" << std::endl;

      if  ( preCandidates.size()==1 ) 
      {
	streamlog_out(DEBUG4) << "Only one preCandidate after regrouping" << std::endl;
	streamlog_out(DEBUG4) << "Original: ";
	for ( std::multiset<SortablePulse>::iterator setIter= pulseSet.begin();
	    setIter !=  pulseSet.end(); ++setIter )
	{
	  streamlog_out(DEBUG4) << setIter->getTrackerPulse()->getCellID0()
				<<"(" 
				<< setIter->getTrackerPulse()->getTime()
				<<")\t ";
	}
	streamlog_out(DEBUG4) << std::endl
			      << "Regrouped: ";
	for (std::vector< SortablePulse >::iterator pulseIter = preCandidates[0].begin();
	     pulseIter < preCandidates[0].end(); pulseIter++)
	{
	  streamlog_out(DEBUG4) << pulseIter->getTrackerPulse()->getCellID0()
				<<"(" 
				<< pulseIter->getTrackerPulse()->getTime()
				<<")\t ";
	}
	streamlog_out(DEBUG4) << std::endl;
      }// if preCandidates.size()==1 
      
      // Do we want single pulses as hit candidates?
      // if yes:
      // * delete all pulses in the set that are already in the pre candidates
      // * make a hit candidate for each pulse left in the set

    }// if regrouping
    else
    {
      // no regrouping, just fill all pulses into one hit candidate
      std::vector< SortablePulse > theOnlyHitCandidate;
      for ( std::multiset<SortablePulse>::iterator setIter= pulseSet.begin();
	    setIter !=  pulseSet.end(); ++setIter )
      {
	theOnlyHitCandidate.push_back( *setIter );
      }
      preCandidates.push_back( theOnlyHitCandidate );
    }

    // run splitter on the regrouped pulses
    for (std::vector< std::vector< SortablePulse > >::iterator preCandidateIter = preCandidates.begin();
	 preCandidateIter < preCandidates.end(); preCandidateIter++)
    {
      // sort the hit candidate
      std::sort( preCandidateIter->begin(), preCandidateIter->end() );
      
      // DEBUG
      
      {
	streamlog_out(DEBUG3) <<"hit from pad "
			      <<	preCandidateIter->begin()->getTrackerPulse()->getCellID0()
			      <<" to pad "
			      <<	preCandidateIter->rbegin()->getTrackerPulse()->getCellID0();
	const gear::TPCModule &module = Global::GEAR->getTPCParameters().getModule( preCandidateIter->rbegin()->getTrackerPulse()->getCellID1()) ;
	int row = module.getRowNumber(preCandidateIter->begin()->getTrackerPulse()->getCellID0());
	streamlog_out(DEBUG3) << " ; nPadsInRow = " 
			      << module.getPadsInRow( row ).size() << std::endl;
      }
	
	

      // fill the hit into a std::vector<float>
      std::vector<float> unseparatedHit;
      // To keep track which signal height came from which pulse 
      // we need to store this information (in case of interpolated pulses there are more entries
      // in the inseparatedHit vector than there are pulses).
      std::vector< std::vector< SortablePulse >::iterator > unseparatedHitReference;
      for (std::vector< SortablePulse >::iterator pulseIter = preCandidateIter->begin();
	   pulseIter != preCandidateIter->end();
	   pulseIter++)
      {
	unseparatedHit.push_back( pulseIter->getTrackerPulse()->getCharge());
	unseparatedHitReference.push_back( pulseIter );

	//fill in interpolated values in case of gaps
	if (pulseIter+1 < preCandidateIter->end() ) // check for gap if not the last entry in vector
	{
	  // the distance between the pad indices
	  int padDistance = ( *(pulseIter+1) - *pulseIter ) ;

	  // if the distance is larger than one there is a gap
	  for (int i=1; i < padDistance ; i++)
	  {
	    unseparatedHit.push_back( pulseIter->getTrackerPulse()->getCharge()*static_cast<float>(padDistance-i) / static_cast<float>(padDistance) 
				      + (pulseIter+1)->getTrackerPulse()->getCharge()*static_cast<float>(i) / static_cast<float>(padDistance) );
	    // fill the end() iterators to indicate it is invalid
	    unseparatedHitReference.push_back(  preCandidateIter->end() );
				     
	  } // distance
	}// if (pulseIter+1 < preCandidateIter->end() )

      }// for pulseIter

      // now the vector of floats for the unseparated hit is filled
      // call the actual splitter
      SignalSplitter splitter;

      std::vector< std::pair< size_t, std::vector<float> > > splitSignals = splitter.splitSignal( unseparatedHit );

      // test if the hit has been split all
      if ( ( splitSignals.size() == 1 ) && (!regroup) )
      {
	streamlog_out(DEBUG4) << "No splitting done" << std::endl;
	// return the empty return vector
	return outputVec;
      }
      else
      {
	// DEBUG: Print the original pulse:
	streamlog_out(DEBUG3) << "signal before splitting: ";
	for (std::vector<float>::iterator unsepIter = unseparatedHit.begin() ; 
	     unsepIter < unseparatedHit.end(); unsepIter++)
	  {
	    streamlog_out(DEBUG3)  << *unsepIter << " ";
	  }
	streamlog_out(DEBUG3) << std::endl;
	

	//create hits from the split signals
	
	for (std::vector< std::pair< size_t, std::vector<float> > >::iterator hitSignalIter = splitSignals.begin();
	     hitSignalIter < splitSignals.end(); hitSignalIter++ )
	{	
	  HitCandidate hitCandidate( Global::GEAR->getTPCParameters(), _vDrift );
	  //	  TrackerHitImpl * newHit =  new TrackerHitImpl;
	  
	  // DEBUG:
	  streamlog_out(DEBUG3) << "Split hit at " << hitSignalIter->first  
				<< " with data " ;
	  for ( size_t index = 0;
		index < hitSignalIter->second.size() ; 
		index ++ )
	    {
	      streamlog_out(DEBUG3) << hitSignalIter->second[index] <<  " ";
	    }
	  streamlog_out(DEBUG3) << std::endl;

	  // fill the pulses into the HitCandidate
	  // Here we use an unsigned int as loop variable 
	  //  since we have to address three vectors with it.
	  
	  for ( size_t index = 0;
		index < hitSignalIter->second.size() ; 
		index ++ )
	  {
//	    sumq += hitSignalIter->second[index];

	    // check that the bin has not been interpolated
	    if ( unseparatedHitReference[index+hitSignalIter->first] != preCandidateIter->end() )
	    {
	      // test whether a pulse has been divided
	      // compare the charge in the split hit with the original charge 
	      TrackerPulse *originalPulse = (*preCandidateIter)[index+hitSignalIter->first].getTrackerPulse();
	      streamlog_out(DEBUG3) << "originalPulse =" << originalPulse << std::endl;
	      if ( hitSignalIter->second[index] != 
		   originalPulse->getCharge() )
	      {
		// pulse has been split, create a new pulse
		TrackerPulseImpl * newPulse = new TrackerPulseImpl;
		newPulse->setCellID0 (originalPulse->getCellID0());
		newPulse->setCellID1 (originalPulse->getCellID1());
		newPulse->setTime (originalPulse->getTime());
		newPulse->setCharge( hitSignalIter->second[index] );
		///\FIXME: set HAS_BEEN_DIVIDED flag
		// newPulse->setQuality (int quality);
		newPulse->setQualityBit (5); /// \FIXME : this is arbitrary

		// leave the tracker data unset? Or do we need even another collection
		//void   setTrackerData (EVENT::TrackerData *corrData)

		_pulsesInSplitHitsCollection->addElement( newPulse );

		hitCandidate.addPulse( newPulse );
	      }
	      else
	      {
		hitCandidate.addPulse( originalPulse );
	      }
	    }// if  unseparatedHitReference[index+hitSignalIter->first] != preCandidateIter->end() 
	  
	  }// for index within split sinal
	  	     
	  TrackerHitImpl * newHit =  hitCandidate.createTrackerHit();
	  
	  // \FIXME !!! correct for negatice z coordinate as long as z in modules 
	  // not implenemted correclty
	  if ( inputHit->getPosition()[2] < 0)
	  {
	    double position[3];
	    
	    // this is a rotation around the y axis
	    position[0] = -newHit->getPosition()[0];
	    position[1] =  newHit->getPosition()[1];
	    position[2] = -newHit->getPosition()[2];

	    newHit->setPosition( position );
	  }

	  // \FIXME !!! replace the quality which has not been set correctly by hitCandiate
	  newHit->setQuality( inputHit->getQuality() ) ;
	  //newHit->setQuality( inputHit->getQuality() | HIT_QUALITY_IS_SPLIT) ;	  
	  
	  // the output collection is declared as a subset. It contains pointers to the
	  // original hit, if they have been used, or to the split ones, but does not store
	  // the hits themselves.
	  // To store the hits they are put into the splitHitsCollection
	  _splitHitsCollection->addElement( newHit );
	  outputVec.push_back( newHit );
	  

	}// for hitSignalIter
	   
      }// else signal has been split
    
    }// for preCandidateIter    

    return outputVec;
  }// for 

bool HitSplitter::SortablePulse::operator<(const HitSplitter::SortablePulse & p) const
{
  const gear::TPCModule &module = Global::GEAR->getTPCParameters().getModule( _pulse->getCellID1() );
  const gear::PadRowLayout2D &  localPadLayout=module.getLocalPadLayout();
  const std::vector < double > &   localPlaneExtent = localPadLayout.getPlaneExtent();

  if ( ( localPadLayout.getCoordinateType() == gear::PadRowLayout2D::POLAR ) &&
       ( fabs( localPlaneExtent[3] - localPlaneExtent[2] - 2*M_PI ) < 1e-6) )
    // do this to check for a full circle geometry, incl. rounding errors in plane extent)
  {
    int thisRowNumber   = localPadLayout.getRowNumber( _pulse->getCellID0() );
    int thisPadNumber   = localPadLayout.getPadNumber( _pulse->getCellID0() );
    int thatRowNumber   = localPadLayout.getRowNumber( p.getTrackerPulse()->getCellID0() );
    int thatPadNumber   = localPadLayout.getPadNumber( p.getTrackerPulse()->getCellID0() );

    if (thisRowNumber!=thatRowNumber)
      throw gear::Exception("SortablePulse::Cannot sort pulses, they are not in the same row!");

    int nPadsInRow = localPadLayout.getPadsInRow( thisRowNumber ).size();

    // This algorithm only works if pulses are not larger than 1/4 of a full circle
    // if thisPulse is in the fourth quadrant and thatPulse is in the first return true
    if ( (thisPadNumber > nPadsInRow*0.75) &&  (thatPadNumber < nPadsInRow*0.25) )
      return true;
    // the other condition we have to catch: this in the first and that in the fourth quadrant
    else if ( (thisPadNumber < nPadsInRow*0.25) &&  (thatPadNumber > nPadsInRow*0.75) )
      return false;
    // for all other cases there is no wrap around, return the normal ordering
    else
      return (thisPadNumber < thatPadNumber);
  }
  else // not a full circle or Cartesian coordinates
    return ( _pulse->getCellID0() < p.getTrackerPulse()->getCellID0() );
}

int HitSplitter::SortablePulse::operator-(const HitSplitter::SortablePulse & p) const
{
  const gear::TPCModule & module = Global::GEAR->getTPCParameters().getModule( _pulse->getCellID1() );
  const gear::PadRowLayout2D &  localPadLayout=module.getLocalPadLayout();
  const std::vector < double > &   localPlaneExtent = localPadLayout.getPlaneExtent();

  if ( ( localPadLayout.getCoordinateType() == gear::PadRowLayout2D::POLAR ) &&
       ( fabs( localPlaneExtent[4] - localPlaneExtent[3] - 2*M_PI ) < 1e-6) )
    // do this to check for a full circle geometry, incl. rounding errors in plane extent)
  {
    int thisRowNumber   = localPadLayout.getRowNumber( _pulse->getCellID0() );
    int thisPadNumber   = localPadLayout.getPadNumber( _pulse->getCellID0() );
    int thatRowNumber   = localPadLayout.getRowNumber( p.getTrackerPulse()->getCellID0() );
    int thatPadNumber   = localPadLayout.getPadNumber( p.getTrackerPulse()->getCellID0() );

    if (thisRowNumber!=thatRowNumber)
      throw gear::Exception("SortablePulse::Cannot sort pulses, they are not in the same row!");

    int nPadsInRow = localPadLayout.getPadsInRow( thisRowNumber ).size();

    // This algorithm only works if pulses are not larger than 1/4 of a full circle
    // if thisPulse is in the fourth quadrant and thatPulse is in the first 
    // take into accout the wrap around
    if ( (thisPadNumber > nPadsInRow*0.75) &&  (thatPadNumber < nPadsInRow*0.25) )
      return -( thatPadNumber + nPadsInRow - thisPadNumber);
    // the other condition we have to catch: this in the first and that in the fourth quadrant
    else if ( (thisPadNumber < nPadsInRow*0.25) &&  (thatPadNumber > nPadsInRow*0.75) )
      return ( thatPadNumber + nPadsInRow - thisPadNumber);
    // for all other cases there is no wrap around, return the normal ordering
    else
      return (thisPadNumber - thatPadNumber);
  }
  else // not a full circle or Cartesian coordinates
    return ( _pulse->getCellID0() - p.getTrackerPulse()->getCellID0() );
}

  HitSplitter::SortablePulse::SortablePulse( EVENT::TrackerPulse * pulse):
    _pulse(pulse)
  {}

  EVENT::TrackerPulse * HitSplitter::SortablePulse::getTrackerPulse() const
  {
    return _pulse;
  }
  
  HitSplitter::Distance::Distance( float d, std::multiset<SortablePulse>::iterator first,
					    std::multiset<SortablePulse>::iterator  second)
    :  distance(d), firstPulse(first), secondPulse(second)
  {}

} // close namespace brackets
