#include "TimePixHighTOTAnalyserProcessor.h"
#include "TimePixPixelMode.h"
#include "TimePixMapHandlerProcessor.h"

// C++
#include <iostream>
#include <cmath>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerRawData.h>
#include <EVENT/TrackerHit.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerRawDataImpl.h>
#include "UTIL/LCTOOLS.h"

// Aida
#ifdef MARLIN_USE_AIDA
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/IHistogram1D.h>
#include <AIDA/IHistogram2D.h>
#endif

//marlin
#include <marlin/ConditionsProcessor.h>

using std::string;
using namespace tpcconddata;
using namespace lcio ;
using namespace marlin ;

namespace marlintpc{


    TimePixHighTOTAnalyserProcessor aTimePixHighTOTAnalyserProcessor ;
    

    TimePixHighTOTAnalyserProcessor::TimePixHighTOTAnalyserProcessor() : Processor("TimePixHighTOTAnalyserProcessor") {
	/* the constructor call first the super constructor which the name 
	 * of the processor
	 * this must by the same as the class name.
	 */
	
	/* description of the processor which will be displayed in the
	 * steering file autometicly generated by 
	 * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
	 */
	_description = "TimePixHighTOTAnalyserProcessor eliminates extremely high TOT values due to a chip fault and replaces them by the arithmetic mean of the surrounding TOT values" ;
	
	
	/* register steering parameters: name, description, class-variable, default value
	 * the type will by definded by the type of the default value
	 * string, double, float and int are possible
	 * plaese note to get a float you must cast in C++: (float) 1.3 
	 */
	registerInputCollection( LCIO::TRACKERRAWDATA,
				 "InputCollectionName" , 
				 "Name of the raw data collection"  ,
				 _RawDataColName ,
				 std::string("TimePixRawData") ) ;

	registerInputCollection( LCIO::LCGENERICOBJECT,
				 "ModeMapCollectionName" , 
				 "Name of the mode map collection"  ,
				 _modeMapColName ,
				 std::string("TimePixModeMap") ) ;

	registerOutputCollection( LCIO::LCGENERICOBJECT,
				 "OutputCollectionName" , 
				 "Name of the HighTOT corrected mode map collection"  ,
				  _modeMapCorrName,
				 std::string("TimePixModeMapCorr") ) ;

	registerProcessorParameter( "NColumn" , 
				    "Number of columns of the chip "  ,
				    _nColumn ,
				    (int) 256 ) ;

	registerProcessorParameter( "NRow" , 
				    "Number of rows of the chip "  ,
				    _nRow ,
				    (int) 256 ) ;

	registerProcessorParameter( "TOTThreshold" , 
				    "threshold on relation of TOT value to neighboring TOT values"  ,
				    _totThreshold ,
				    (double) 300 ) ;
	
	registerProcessorParameter( "SetOutputTransient" , 
				    "if not 0 the output collection is set transient"  ,
				    _outputIsTransient,
				    0);

	_chipID = 0;

	_timePixMap = 0;
	_currentTimePixMap = 0;
	_correctedTimePixMap = 0;
	
	_mapHandlerProcessor = 0;

    }
    
    
    void TimePixHighTOTAnalyserProcessor::init() { 

      streamlog_out(DEBUG2)<<"TimePixHighTOTAnalyserProcessor::init()"<<std::endl;

	// usually a good idea to
	printParameters() ;

	_chipID = 0;

	//register this processor with the TimePixMapHandlerProcessor
	_mapHandlerProcessor = TimePixMapHandlerProcessor::registerHighTOTProcessor( this );

	if ( _mapHandlerProcessor == 0 ) 
	    throw Exception("TimePixHighTOTAnalyserProcessor::init: "
			    "TimePixMapHandlerProcessor is not initialised - without it, "
			    "there is no TimePixMap needed by TimePixHighTOTAnalyserProcessor!");

	//define histograms
	highTOT =   AIDAProcessor::histogramFactory(this)->
	  createHistogram1D( "highTOT", "010707_001.siy.300.310 TOT value", 24000, -12000, 12000);

	TOTTOTCorrelation =   AIDAProcessor::histogramFactory(this)->
	  createHistogram1D( "TOTTOTCorrelation", 
			     "010707_001.siy.300.310 Correlation between high TOT values and neighboring TOT values", 
			     2000, 0, 2000);

	TimeTOTCorrelation =   AIDAProcessor::histogramFactory(this)->
	  createHistogram1D( "TimeTOTCorrelation", 
			     "010707_001.siy.300.310 Correlation between high TOT values and neighboring Time values", 
			     2000, -1000, 1000);

	
    }
    
    void TimePixHighTOTAnalyserProcessor::processRunHeader( LCRunHeader* run) { 
	
	streamlog_out(DEBUG2)<<"TimePixHighTOTAnalyserProcessor::processRunHeader("
			     <<run->getRunNumber()<<")"<<std::endl;
	
	run->parameters().setValue( _processorName + "_revision", "$Rev: 1873 $");
	
	for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
	    if ( ! i->second->isOptional() || i->second->valueSet() ) {
		run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
	    }
	}
	
	
    } 
    
    void TimePixHighTOTAnalyserProcessor::processEvent( LCEvent * evt ) { 
      
	streamlog_out(MESSAGE1)<<"Event "<<evt->getEventNumber()<<std::endl;
	
	streamlog_out(DEBUG2)<<"TimePixHighTOTAnalyserProcessor::processEvent( "
			     <<evt->getEventNumber()<<" ) "<<std::endl;
	
	streamlog_out(DEBUG0)<<"collections in event"<<std::endl;
	const std::vector<std::string > * ColNames = evt->getCollectionNames();
	for ( unsigned int i = 0; i < ColNames->size(); i++ ) {
	
	    streamlog_out(DEBUG0)<<(*ColNames)[i]<<std::endl;
	}
      
      
	//get the current TimePixMap from the MapHandlerProcessor
	_timePixMap = _mapHandlerProcessor->getUnmodifiedTimePixMap();
	
	
	// get the input collection from the event
	LCCollection* rawDataCol = 0;
	try { rawDataCol = evt->getCollection( _RawDataColName ) ; }
	
	catch(DataNotAvailableException){
	    // if there is no input collection give a warning to the user
	    std::cout<<"Warning of TimePixHighTOTAnalyserProcessor::processEvent: input collection "
		     <<_RawDataColName<<" cannot be found in event "<<evt->getEventNumber()<<std::endl;	
	}  
	

	LCCollectionVec* modeMapCorr = new LCCollectionVec(LCIO::LCGENERICOBJECT ); 
	
	// get the number of elements in the input collection, for TimePixHit usually 1
	int nRawData = rawDataCol->getNumberOfElements();
	//and loop over it
	for(int iRawData=0; iRawData< nRawData ; iRawData++){
	    
	    /* getElement returns only a LCObject
	     * you have to cast it to the get the right type
	     * please use dynamic_cast which will return 0 if the object cannot be casted
	     */
	    TrackerRawData* RawData = 
		dynamic_cast<TrackerRawData*>( rawDataCol->getElementAt( iRawData ) ) ;
	    
	    if ( RawData != 0 ) {
		//only proceed if the cast was successful
		
		int cellid0 = RawData->getCellID0();
		int x = cellid0%_nColumn;
		int y = static_cast<int>((cellid0-x)/_nColumn); 
		
		ShortVec ADCValVec = RawData->getADCValues();
		
		streamlog_out(DEBUG0)<<"ADCValVec.size() = "<<ADCValVec.size()<<std::endl;
		
		ShortVec::iterator iADCVal;
		
		//-----------------------------
		for ( iADCVal = ADCValVec.begin(); iADCVal != ADCValVec.end(); iADCVal++ ) {
		    
		    
		    if ( *iADCVal < 0 ) {
			
			modeMapCorr->addElement(new TimePixPixelMode( RawData->getCellID1(), 
								      cellid0, 
								      TimePixPixelMode::ONEHIT));
			
			
			
			cellid0++;
			x++;
			if ( x == _nColumn ) {
			    x = 0;
			    y++;
			}
			
		    }
		    
		    else {
			
			//determine mode
			int mode = ((_timePixMap->find( TimePixPixelMode::PixelKey( _chipID, cellid0 ) ))
				    ->second)->getPixelMode();
			if ( mode == TimePixPixelMode::TOT ) {
			    
			    if ( iADCVal < ADCValVec.end()-1 && *(iADCVal+1) > 0 ) {
				
				//test if this following pixel is also TOT
				int mode1 = 
				    ((_timePixMap->find( TimePixPixelMode::PixelKey( _chipID, cellid0 ) ))
				     ->second)->getPixelMode();
				
				if ( mode1 == TimePixPixelMode::TOT && 
				     mode == TimePixPixelMode::TOT ) { 
				    //if yes compare those two pixels
				    
				    double ratio = ADCValVec[cellid0]/ADCValVec[cellid0+1];
				    if ( ratio > 300 )   modeMapCorr->addElement
							     ( new TimePixPixelMode
							       ( RawData->getCellID1(), 
								 cellid0, 
								 TimePixPixelMode::ONEHIT ) );
				    
				}
				
				else { //try next pixel
		    
				    if ( iADCVal+1 < ADCValVec.end()-1 && 
					 *(iADCVal+2) > 0 ) {
					
					//test if the next next pixel is also TOT
					int mode2 = 
					    ((_timePixMap->find( TimePixPixelMode::PixelKey( _chipID, cellid0 ) ))
					     ->second)->getPixelMode();
					
					if ( mode2 == TimePixPixelMode::TOT && 
					     mode == TimePixPixelMode::TOT ) { 
					    //if yes compare those two pixels
					    
					    double ratio = ADCValVec[cellid0]/ADCValVec[cellid0+2];
					    if ( ratio > 300 )   
						modeMapCorr->addElement
						    ( new TimePixPixelMode( RawData->getCellID1(), 
									    cellid0, 
									    TimePixPixelMode::ONEHIT ) );
			   
					}
					
				    } //end of if ( iADCVal+1 < ADCValVec.end()-1 && *(iADCVal+2) > 0 ) {
				    
		    
				}//end of else
				
				
			    }// end of if ( *iADCVal > 0 && iADCVal < ADCValVec.end()-1 && *(iADCVal+1) > 0 ) {
			    
			} //end of if ( mode == TimePixPixelMode::TOT ) {
			
			
			cellid0++;
			x++;
			if ( x == _nColumn ) {
			    x = 0;
			    y++;
			}
			
			
		    }// end of for ( iADCVal = ADCValVec.begin(); iADCVal != ADCValVec.end(); iADCVal++ ) {
		    
		}// end of if ( RawData != 0 ) {
		
	    } //end of else
	    
	}// end of for(int iRawData=0; iRawData< nRawData ; iRawData++){

	if (isFirstCollection){
	    modeMapCorr->parameters().setValue("RawDataCollectionName",_RawDataColName);
	    // TO DO: does this make sence? 
	    modeMapCorr->parameters().setValue("SetOutputTransient",_outputIsTransient);
	    isFirstCollection = false;
	}
	
	if ( modeMapCorr->getNumberOfElements() != 0 ) evt->addCollection( modeMapCorr, _modeMapCorrName );
      
      
      	
	//now update _timePixMap
	updateTimePixModeMap( evt );
	
      
	streamlog_out(DEBUG2)<<"end of TimePixHighTOTAnalyserProcessor::checkEvent( "
			     <<evt->getEventNumber()<<" )"<<std::endl;
      

    }
    
    
    void TimePixHighTOTAnalyserProcessor::check( LCEvent * evt ) { 
      
	// get the input collection from the event
	LCCollection* rawDataCol = 0;
	try { rawDataCol = evt->getCollection( _RawDataColName ) ;}
	catch(DataNotAvailableException){
	    // if there is not input collection give a warning to the user
	    std::cout<<"Warning of TimePixHighTOTAnalyserProcessor::processEvent: input collection "
		     <<_RawDataColName<<" cannot be found in event "<<evt->getEventNumber()<<std::endl;	
	} 
	
	
	// get the number of elements in the input collection, for TimePixHit usually 1
	int nRawData = rawDataCol->getNumberOfElements();
	//and loop over it
	for(int iRawData=0; iRawData< nRawData ; iRawData++){
	    
	    /* getElement returns only a LCObject
	     * you have to cast it to the get the right type
	     * please use dynamic_cast which will return 0 if the object cannot be casted
	     */
	    TrackerRawData* RawData = 
		dynamic_cast<TrackerRawData*>( rawDataCol->getElementAt( iRawData ) ) ;
	    
	    if ( RawData != 0 ) {
		//only proceed if the cast was successful
		
		int cellid0 = RawData->getCellID0();
		int x = cellid0%_nColumn;
		int y = static_cast<int>((cellid0-x)/_nColumn); 
		
		ShortVec ADCValVec = RawData->getADCValues();
		
		ShortVec::iterator iADCVal;
		for ( iADCVal = ADCValVec.begin(); iADCVal != ADCValVec.end(); iADCVal++ ) {
		    
		    int mode = -1;
		    
		    int mode_1 = -1;
		    int mode_2 = -1;
		    int mode_3 = -1;
		    int mode_4 = -1;
		    int mode_5 = -1;
		    int mode_6 = -1;
		    int mode_7 = -1;
		    int mode_8 = -1;
		
		    int cellid_1 = cellid0 - _nColumn - 1;
		    int cellid_2 = cellid0 - _nColumn;
		    int cellid_3 = cellid0 - _nColumn + 1;
		    int cellid_4 = cellid0 - 1;
		    int cellid_5 = cellid0 + 1;
		    int cellid_6 = cellid0 + _nColumn - 1;
		    int cellid_7 = cellid0 + _nColumn;
		    int cellid_8 = cellid0 + _nColumn + 1;
		    
		    //left border
		    if ( cellid_1%_nColumn == _nColumn - 1 ) cellid_1 = -1;
		    if ( cellid_4%_nColumn == _nColumn - 1 ) cellid_4 = -1;
		    if ( cellid_6%_nColumn == _nColumn - 1 ) cellid_6 = -1;
		    
		    //right border
		    if ( cellid_3%_nColumn == 0 ) cellid_3 = -1;
		    if ( cellid_5%_nColumn == 0 ) cellid_5 = -1;
		    if ( cellid_8%_nColumn == 0 ) cellid_8 = -1;
		    
		    //end of chip
		    if ( cellid_1 >= _nColumn*_nRow ) cellid_1 = -1;
		    if ( cellid_2 >= _nColumn*_nRow ) cellid_2 = -1;
		    if ( cellid_3 >= _nColumn*_nRow ) cellid_3 = -1;
		    if ( cellid_4 >= _nColumn*_nRow ) cellid_4 = -1;
		    if ( cellid_5 >= _nColumn*_nRow ) cellid_5 = -1;
		    if ( cellid_6 >= _nColumn*_nRow ) cellid_6 = -1;
		    if ( cellid_7 >= _nColumn*_nRow ) cellid_7 = -1;
		    if ( cellid_8 >= _nColumn*_nRow ) cellid_8 = -1;
		    
		    try{
			mode = 
			    ((_timePixMap->find( TimePixPixelMode::PixelKey( _chipID, cellid0 ) ))
			     ->second)->getPixelMode();
		    }
		    catch(Exception& e ){
			std::cout << " Exception: " << e.what() << std::endl ;
		    }
		    
		    if ( cellid_1 > 0 && cellid_1 < _nColumn*_nRow ) {
			
			try{
			    mode_1 = 
				((_timePixMap->find( TimePixPixelMode::PixelKey( _chipID, cellid0 ) ))
				 ->second)->getPixelMode();

			}
			catch(Exception& e ){
			    std::cout << " Exception: " << e.what() << std::endl ;
			}
			
		    }
		    
		    if ( cellid_2 > 0 && cellid_2 < _nColumn*_nRow ) {
			
			try{
			    mode_2 = 
				((_timePixMap->find( TimePixPixelMode::PixelKey( _chipID, cellid0 ) ))
				 ->second)->getPixelMode();
			}
			catch(Exception& e ){
			    std::cout << " Exception: " << e.what() << std::endl ;
			}
			
		    }
		    
		    if ( cellid_3 > 0 && cellid_3 < _nColumn*_nRow ) {
			
			try{
			    mode_3 = 
				((_timePixMap->find( TimePixPixelMode::PixelKey( _chipID, cellid0 ) ))
				 ->second)->getPixelMode();
			}
			catch(Exception& e ){
			    std::cout << " Exception: " << e.what() << std::endl ;
			}
			
		    }
		    
		    if ( cellid_4 > 0 && cellid_4 < _nColumn*_nRow ) {
			
			try{
			    mode_4 = 
				((_timePixMap->find( TimePixPixelMode::PixelKey( _chipID, cellid0 ) ))
				 ->second)->getPixelMode();
			}
			catch(Exception& e ){
			    std::cout << " Exception: " << e.what() << std::endl ;
			}
			
		    }
		    
		    if ( cellid_5 > 0 && cellid_5 < _nColumn*_nRow ) {
			
			try{
			    mode_5 = 
				((_timePixMap->find( TimePixPixelMode::PixelKey( _chipID, cellid0 ) ))
				 ->second)->getPixelMode();
			}
			catch(Exception& e ){
			    std::cout << " Exception: " << e.what() << std::endl ;
			}
			
		    }
		    
		    if ( cellid_6 > 0 && cellid_6 < _nColumn*_nRow ) {
			
			try{
			    mode_6 = 
				((_timePixMap->find( TimePixPixelMode::PixelKey( _chipID, cellid0 ) ))
				 ->second)->getPixelMode();
			}
			catch(Exception& e ){
			    std::cout << " Exception: " << e.what() << std::endl ;
			}
			
		    }
		    
		    if ( cellid_7 > 0 && cellid_7 < _nColumn*_nRow ) {
			
			try{
			    mode_7 = 
				((_timePixMap->find( TimePixPixelMode::PixelKey( _chipID, cellid0 ) ))
				 ->second)->getPixelMode();
			}
			catch(Exception& e ){
			    std::cout << " Exception: " << e.what() << std::endl ;
			}
			
		    }
		    
		    if ( cellid_8 > 0 && cellid_8 < _nColumn*_nRow ) {
			
			try{
			    mode_8 = 
				((_timePixMap->find( TimePixPixelMode::PixelKey( _chipID, cellid0 ) ))
				 ->second)->getPixelMode();
			}
			catch(Exception& e ){
			    std::cout << " Exception: " << e.what() << std::endl ;
			}
			
		    }
		    
		    if ( mode == TimePixPixelMode::TOT && *iADCVal > 0 ) {
			
			if ( mode_1 == TimePixPixelMode::TOT && ADCValVec[cellid_1] != 0 ) 
			    TOTTOTCorrelation->fill( ADCValVec[cellid_1] / ADCValVec[cellid0]  );

			if ( mode_3 == TimePixPixelMode::TOT && ADCValVec[cellid_3] != 0 ) 
			    TOTTOTCorrelation->fill( ADCValVec[cellid_3] / ADCValVec[cellid0]  );

			if ( mode_6 == TimePixPixelMode::TOT && ADCValVec[cellid_6] != 0 ) 
			    TOTTOTCorrelation->fill( ADCValVec[cellid_6] / ADCValVec[cellid0]  );

			if ( mode_8 == TimePixPixelMode::TOT && ADCValVec[cellid_8] != 0 ) 
			    TOTTOTCorrelation->fill( ADCValVec[cellid_8] / ADCValVec[cellid0]  );
			
			
			if ( mode_2 == TimePixPixelMode::TIMEPIX && ADCValVec[cellid_2] != 0 ) 
			    TimeTOTCorrelation->fill( ADCValVec[cellid_2] - ADCValVec[cellid0]  );

			if ( mode_4 == TimePixPixelMode::TIMEPIX && ADCValVec[cellid_4] != 0 ) 
			    TimeTOTCorrelation->fill( ADCValVec[cellid_4] - ADCValVec[cellid0]  );

			if ( mode_5 == TimePixPixelMode::TIMEPIX && ADCValVec[cellid_5] != 0 ) 
			    TimeTOTCorrelation->fill( ADCValVec[cellid_5] - ADCValVec[cellid0]  );

			if ( mode_7 == TimePixPixelMode::TIMEPIX && ADCValVec[cellid_7] != 0 ) 
			    TimeTOTCorrelation->fill( ADCValVec[cellid_7] - ADCValVec[cellid0]  );
			
		    }
		    
		    if ( mode == TimePixPixelMode::TOT  ) highTOT->fill( *iADCVal );
			
		    

		
		    cellid0++;
		    x++;
		    if ( x == _nColumn ) {
			x = 0;
			y++;
		    }
		    
		} //end of for ( iADCVal = ADCValVec.begin(); iADCVal != ADCValVec.end(); iADCVal++ ) {
		
		ADCValVec.clear();
		
		streamlog_out(DEBUG1)<<"end of if ( RawData != 0 )"<<std::endl;
		
	    } //end of if ( RawData != 0 ) {
	    
	    
	} //end of for(int iRawData = 0; ...
	


	streamlog_out(DEBUG2)<<"end of TimePixHighTOTAnalyserProcessor::checkEvent( "<<evt->getEventNumber()<<" )"<<std::endl;
	
    }

    
    
    void TimePixHighTOTAnalyserProcessor::end(){ 
      /*	
      streamlog_out(MESSAGE4) << "TimePixHighTOTAnalyserProcessor::end()  " << name() 
		<< " processed " << evt->getEventNumber() << " events in " << evt->getRunNumber() << " runs "
		<< std::endl ;
      */	
    }
    
    void TimePixHighTOTAnalyserProcessor::updateTimePixModeMap( LCEvent* evt ){
	
	bool tobecorrected = false;
	
	const std::vector<std::string > * ColNames = evt->getCollectionNames();
	for ( unsigned int i = 0; i < ColNames->size(); i++ ) {
	    
	    std::string colname = (*ColNames)[i];
	    
	    if( colname == _modeMapCorrName ) {
		//correction to mode map exists, now update the mode map
		streamlog_out(DEBUG4)<<"correction to mode map exists, now update the mode map"<<std::endl;
		tobecorrected = true;
		break;
		
	    }
	    
	}
	
	if ( tobecorrected ) {
	    
	    delete _correctedTimePixMap;
	    _correctedTimePixMap = new tpcconddata::TimePixMap();
	    
	    LCCollection* CorrModeMapCol = evt->getCollection( _modeMapCorrName );
	    for ( int i = 0; i < CorrModeMapCol->getNumberOfElements(); i++ ) {

		TimePixPixelMode *pm = 
		    new TimePixPixelMode(dynamic_cast<LCGenericObject*>( CorrModeMapCol->getElementAt( i )));
    
		( (*_correctedTimePixMap)[ pm->getPixelKey() ] )->setPixelMode( pm->getPixelMode() );
		
		
	    }
	    
	    _currentTimePixMap = _correctedTimePixMap;
	    
	}
	else _currentTimePixMap = _timePixMap;
	
	
    }
    
    tpcconddata::TimePixMap* TimePixHighTOTAnalyserProcessor::getUpdatedTimePixMap(){

	//returns the _correctedTimePixMap if it has been corrected
        //else it returns the unmodified _timePixMap
	return _currentTimePixMap;

	
    }
    
}



