#include "TrackFinderHoughTrafoProcessor.h"

// C++
#include <iostream>
#include <vector>
#include <string>
#include <sstream>
#include <cmath>
#include <algorithm>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerHit.h>
#include <EVENT/Track.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerHitImpl.h>
#include <IMPL/TrackImpl.h>
#include <UTIL/LCTOOLS.h>
#include <IMPL/LCFlagImpl.h>

// Aida
#ifdef MARLIN_USE_AIDA
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>

//marlin
#include <marlin/ConditionsProcessor.h>


#endif

using std::string;
//using namespace tpcconddata;
using namespace lcio ;
using namespace marlin ;

namespace marlintpc{

    TrackFinderHoughTrafoProcessor aTrackFinderHoughTrafoProcessor ;
    

    TrackFinderHoughTrafoProcessor::TrackFinderHoughTrafoProcessor() : 
	Processor("TrackFinderHoughTrafoProcessor") , histod(0) , HistoEvent(0)
    {
	/* the constructor calls first the super constructor which is the name 
	 * of the processor
	 * this must by the same as the class name.
	 */
	
	/* description of the processor which will be displayed in the
	 * steering file autometicly generated by 
	 * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
	 */
      _description = "TrackFinderHoughTrafoProcessor uses the hough transformation to find hits on track, track parameters can be used as starting values for the TrackFitter";
          	
	
	/* register steering parameters: name, description, class-variable, default value
	 * the type will by definded by the type of the default value
	 * string, double, float and int are possible
	 * plaese note to get a float you must cast in C++: (float) 1.3 
	 */
	registerInputCollection( LCIO::TRACKERHIT,
				 "InputCollectionName" , 
				 "Name of the input collection (default: TimePixHits)"  ,
				 _inputColName ,
				 std::string("TimePixHits") ) ;
	
	registerOutputCollection( LCIO::TRACK,
				  "OutputCollectionName" , 
				  "Name of the output collection (default: TimePixTrackCandidates)"  ,
				  _outputColName ,
				  std::string("TimePixTrackCandidates") ) ;

	registerProcessorParameter( "TrackType" , 
				    "Type of track that is looked for",
				    _trackType ,
				    std::string("StraightLine") ) ;

	registerProcessorParameter( "SetOutputTransient" , 
				    "if not 0 the output collection is set transient (default 0)"  ,
				    _outputIsTransient,
				    0);

	registerProcessorParameter( "NHitsOnTrack" , 
				    "minimum number of hits on one track"  ,
				    _nHitsOnTrack,
				    5);

	registerProcessorParameter( "MaximumDistanceToTrack" , 
				    "Maximum distance of a hit to the track in mm"  ,
				    _maxDistanceToTrack,
				    0.5);

	registerProcessorParameter( "HoughSpaceNBinsA" , 
				    "The number of bins in the hough space histogram in 'a' direction (default 200)"  ,
				    _houghSpaceNBinsA,
				    200);

	registerProcessorParameter( "HoughSpaceNBinsB" , 
				    "The number of bins in the hough space histogram in 'b' direction (default 120)"  ,
				    _houghSpaceNBinsB,
				    120);

	registerProcessorParameter( "HoughSpaceMaxA" , 
				    "The half range (-max..max) of the hough space histogram in 'a' direction (default 1.2)"  ,
				    _houghSpaceMaxA,
				    float(1.2));

	registerProcessorParameter( "HoughSpaceMaxB" , 
				    "The half range (-max..max) of the hough space histogram in 'b' direction (default 20)"  ,
				    _houghSpaceMaxB,
				    float(20.));

	_chipID = 0;

    }
    
    
    void TrackFinderHoughTrafoProcessor::init() { 
	
	// usually a good idea to
	printParameters() ;

	histod = AIDAProcessor::histogramFactory(this)->
	  createHistogram1D( "histod", "distance center of gravity to track", 
			     40 , -10, 10 ); 


    }
    
    void TrackFinderHoughTrafoProcessor::processRunHeader( LCRunHeader* run) { 
	

      run->parameters().setValue( _processorName + "_revision", "$Rev: 1726 $");

      for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
	if ( ! i->second->isOptional() || i->second->valueSet() ) {
	  run->parameters().setValue( _processorName + "_" + i->second->name(), 
				      i->second->value() );
	}
      }

					       

    } 
    
    void TrackFinderHoughTrafoProcessor::processEvent( LCEvent * evt ) { 

      streamlog_out(MESSAGE1)<<"Event "<<evt->getEventNumber()<<std::endl;

      streamlog_out(DEBUG0)<<"processEvent( "<<evt->getEventNumber()<<")"<<std::endl;


      // get the input collection from the event
      LCCollection* inputCol=0;
      try{      
	inputCol = evt->getCollection( _inputColName ) ;
      }

      catch(DataNotAvailableException){
	// if there is not input collection give a warning to the user
	streamlog_out(WARNING2)
	  <<"processEvent: input collection "
	  <<_inputColName<<" cannot be found in event "<<evt->getEventNumber()<<std::endl;
	return;
      }



      // create the  output collections with a string defining the TYPE (see lcio.h)
      LCCollectionVec* outputCol = new LCCollectionVec( LCIO::TRACK ); 

      // add hits to track collection
      LCFlagImpl trkFlag(0) ;
      trkFlag.setBit( LCIO::TRBIT_HITS ) ;
      outputCol->setFlag( trkFlag.getFlag() );

	
      // set the transient flag as definded by the processor parameter
      outputCol->setTransient( (_outputIsTransient != 0) );

      //vector of trackcandidates
      std::vector<TrackImpl*> TrackCandidateVec;
	


      //set parameters
//      int houghSpaceBinX = 200;//200;
//      int houghSpaceBinY = 120;//120;
//      float houghSpaceMaxX = 1.2;
//      int houghSpaceMaxY = 20;

      std::stringstream sEvt;
      sEvt << evt->getEventNumber();
 
      //intialize vectors
      TPVec maxVec;
      TPVec foundVec;

      
      int histonumber = 0;

      for ( std::vector<AIDA::IHistogram2D*>::iterator i = HistoHoughVec.begin();
	    i < HistoHoughVec.end(); i++ ) {
	
	AIDAProcessor::histogramFactory(this)->destroy( *i );	

      }


      HistoHoughVec.clear();

      //loop over orientation of axis 
      //j = 0: xy   //j = 1: yx

      for ( int j = 0; j < 2; j ++ ) {

	streamlog_out(DEBUG0)<<"fill inputColVec"<<std::endl;
	//create a copy of the inputCollection that can be written on

	//fill copy of inputcollection	
	std::vector<TrackerHit*> inputColVec;
	int nInput = inputCol->getNumberOfElements();
	for ( int iInput = 0; iInput < nInput; iInput++) {
	  TrackerHit* Hit = 
	    dynamic_cast<TrackerHit*>( inputCol->getElementAt( iInput ) ) ;
	  if ( Hit != 0 ) {
	    
	    inputColVec.push_back( Hit );
	    
	  }
	}
	
	streamlog_out(DEBUG0)<<"filled inputColVec : "<<inputColVec.size()<<std::endl;
      
	while ( inputColVec.size() != 0 ) {
	
	  std::stringstream shistonumber;
	  shistonumber << histonumber;
	  std::string orientation;
	  if ( j == 0 ) orientation = "_xy";
	  if ( j == 1 ) orientation = "_yx";
	  std::string histoname = "HistoHough" + sEvt.str() + "_" + shistonumber.str() + orientation;

	  streamlog_out(DEBUG0)<<"creating histogram : "<<histoname<<std::endl;
	  
	  HistoHoughVec.push_back( AIDAProcessor::histogramFactory(this)->
				   createHistogram2D( histoname.c_str(), "Hough space", 
						      _houghSpaceNBinsA , -_houghSpaceMaxA, _houghSpaceMaxA, 
						      _houghSpaceNBinsB , -_houghSpaceMaxB, _houghSpaceMaxB) );
	  
	  AIDA::IHistogram2D *HistoHough = *(HistoHoughVec.end()-1);

	  for( std::vector<TrackerHit*>::iterator iHit = inputColVec.begin();
	       iHit < inputColVec.end(); iHit++ ) {
	    
	    bool removed = false;
	    
	    TrackerHit* RawHit = *iHit;
	    
	    if( RawHit != 0 ) {
	      
	      const double *pos = RawHit->getPosition(); //hit position x, y, z in mm
	      
	      //test if point already belongs to a track
	      //..........................................
	      
	      for ( TPVec::iterator itFoundVec = foundVec.begin();
		    itFoundVec != foundVec.end(); itFoundVec++ ) {
		
		if ( j == 0 &&
		     pos[1] > ( (*itFoundVec).a * pos[0] + (*itFoundVec).b - _maxDistanceToTrack) &&
		     pos[1] < ( (*itFoundVec).a * pos[0] + (*itFoundVec).b + _maxDistanceToTrack) ){
		  //point already belongs to a track and should be deleted
		  inputColVec.erase( iHit );
		  iHit--;
		  removed = true;

		}
	      
		if ( j == 1 &&
		     pos[0] > ( (*itFoundVec).a * pos[1] + (*itFoundVec).b - _maxDistanceToTrack) &&
		     pos[0] < ( (*itFoundVec).a * pos[1] + (*itFoundVec).b + _maxDistanceToTrack) ){
		  //point already belongs to a track and should be deleted
		  inputColVec.erase( iHit );
		  iHit--;
		  removed = true;

		}
 

	      } //for ( std::vector<DoubleVec>::iterator itFoundVec = foundVec.begin();

	      if ( !removed && inputColVec.size() != 0 ) { 
		//if point has not been removed fill in HistoHough
		
		if ( _trackType == "StraightLine" ) { //y = ax + b => b = y - ax
		  
		    for ( float a = -_houghSpaceMaxA; a < _houghSpaceMaxA; 
			a+=2*_houghSpaceMaxA/_houghSpaceNBinsA) {
		    
		    if ( j == 0 ) HistoHough->fill( a, pos[1] - a*pos[0] );
		    if ( j == 1 ) HistoHough->fill( a, pos[0] - a*pos[1] );
		    
		  }
		  
		  
		} //if ( _trackType == "StraightLine" )
		
	      } //if ( !removed )
	      
	    } //end of if ( RawHit != 0 ) 
	    
	  } //end of for(int iHit=0; iHit< nHit ; iHit++){
	  
	  
	  //find the position of the maximum in the Hough histogram
	  //========================================================
	  if ( HistoHough->entries() > 0 ) {
	    
	    streamlog_out(DEBUG0)<<"HistoHough->entries() > 0 : \ninputColVec.size() = "
				 <<inputColVec.size()<<"\n-----------"<<std::endl;
	    
	    //DoubleVec maxBin = maxBinFinder( HistoHough );
	    TrackFinderHoughTrafoProcessor::TrackProperties maxBin = maxBinFinder( HistoHough );

	    int hitsontrack = 0;
	    for( std::vector<TrackerHit*>::iterator iHit = inputColVec.begin();
		 iHit < inputColVec.end(); iHit++ ) {
	      
	      TrackerHit* RawHit = *iHit;
	      
	      if( RawHit != 0 ) {
		
		const double *pos = RawHit->getPosition(); //hit position x, y, z in mm
		
		if ( j == 0 && //xy
		     pos[1] > ( maxBin.a * pos[0] + maxBin.b - _maxDistanceToTrack) &&
		     pos[1] < ( maxBin.a * pos[0] + maxBin.b + _maxDistanceToTrack) ){
		  
		  hitsontrack++;
		  
		}
		
		if ( j == 1 && //yx
		     pos[0] > ( maxBin.a * pos[1] + maxBin.b - _maxDistanceToTrack) &&
		     pos[0] < ( maxBin.a * pos[1] + maxBin.b + _maxDistanceToTrack) ){
		  
		  hitsontrack++;
		  
		}
	      }
	    }
	    
	    maxBin.hitsontrack = hitsontrack;
	    maxBin.orientation = j;
	    foundVec.push_back( maxBin );
		    
	    //only if there are at least _nHitsOnTrack points on the track 
	    //it really is a track
	    if ( hitsontrack > _nHitsOnTrack || hitsontrack == _nHitsOnTrack ) 
	      maxVec.push_back( maxBin );
	    
	    
	  }
	  else {
	    
	    streamlog_out(DEBUG0)<<"HistoHough->entries() !> 0 : break!!!"<<std::endl; 
	    break;
	  }
	  
	  histonumber++;
	  
	  streamlog_out(DEBUG0)<<"end of while: inputColVec.size() = "
			       <<inputColVec.size()<<std::endl;
	  


	} // while ( inputColVec.size() != 0 )  
	
	streamlog_out(DEBUG0)<<"after while: inputColVec.size() = "
			     <<inputColVec.size()<<std::endl;


	inputColVec.clear();
	foundVec.clear();
	
	streamlog_out(DEBUG0)<<"maxVec.size() = "<<maxVec.size()<<std::endl;


      } //end of loop over orientation of axis

     
      //sort maxVec by number of hits on track:
      //=========================================
      std::sort( maxVec.begin(), maxVec.end(), sortPredicate );

      
      streamlog_out(DEBUG0)<<"hits on track sorted:\n.............."<<std::endl;

      
      for ( TPVec::iterator itMaxVec = maxVec.begin();
	    itMaxVec != maxVec.end(); itMaxVec++ ) {
	
	streamlog_out(DEBUG0)<<(*itMaxVec).hitsontrack<<std::endl;
	
      }
      
      
      
      
      //fill vector with track candidates
      //===================================
      
      
      for ( TPVec::iterator itMaxVec = maxVec.begin();
	    itMaxVec != maxVec.end(); itMaxVec++ ) {
	
	
	TrackImpl* trackCandidate = new TrackImpl();
	
	if ( (*itMaxVec).orientation == 0 ) {
	  trackCandidate->setPhi( atan( (*itMaxVec).a ) );
	  trackCandidate->setD0( (*itMaxVec).b * cos( trackCandidate->getPhi() ) );
	}
	
	if ( (*itMaxVec).orientation == 1 ) {
	  trackCandidate->setPhi( M_PI/2 - atan( (*itMaxVec).a ) );
	  trackCandidate->setD0( -((*itMaxVec).b/(*itMaxVec).a)*sin(trackCandidate->getPhi()) );
	}
	
	
	TrackCandidateVec.push_back( trackCandidate );
	
      }


      //FILL OUTPUT COLLECTION
      //==========================

      //every center of gravity is added to the first track that is in reach
      //tracks are sorted by number of possible hits
      for(int iHit=0; iHit < inputCol->getNumberOfElements() ; iHit++){
	    
	TrackerHit* RawHit = 
	  dynamic_cast<TrackerHit*>( inputCol->getElementAt( iHit ) ) ;
	
	if( RawHit != 0 ) {
	  
	  const double *pos = RawHit->getPosition(); //hit position x, y, z in mm
	  
	  //add hit to track if distance to track is < _maxDistanceToTrack
	  for ( std::vector<TrackImpl*>::iterator iTrack = TrackCandidateVec.begin();
		iTrack < TrackCandidateVec.end(); iTrack++ ) {
	    
	    DoubleVec par;
	    par.push_back( tan( (*iTrack)->getPhi() ) );
	    par.push_back( (*iTrack)->getD0() );
	    double d = distanceToTrackFinder( pos, par, "StraightLine"  );
	    if ( fabs(d) < fabs(_maxDistanceToTrack) ) {


	      streamlog_out(DEBUG0)<<"d = "<<d<<" < "<<_maxDistanceToTrack<<std::endl;
	      (*iTrack)->addHit( RawHit );
	      break;

	    }
	    
	  }
	  
	  
	  
	  
	} //end of if ( RawHit != 0 ) 
	
      } //end of for(int iHit=0; iHit< nHit ; iHit++){
      


	//add the tracks to the output collection
      for ( std::vector<TrackImpl*>::iterator iTrack = TrackCandidateVec.begin();
	    iTrack < TrackCandidateVec.end(); iTrack++ )  {
	
	  streamlog_out(DEBUG0)<<"Track consists of "
			       <<((*iTrack)->getTrackerHits()).size()<<" hits"<<std::endl;
	  
	  if ( ((*iTrack)->getTrackerHits()).size() >   static_cast<unsigned int>(_nHitsOnTrack) || 
	       ((*iTrack)->getTrackerHits()).size() ==  static_cast<unsigned int>(_nHitsOnTrack)  ) {
	      
	      outputCol->addElement( *iTrack );
	      
	  } //end of if(((*iTrack)->getTrackerHits()).size()>static_cast<unsigned int>(_nHitsOnTrack)...

	  else {
	      delete *iTrack;
	      *iTrack=0;
	  }
	  //Don't use the TrackCandidateVec after this loop, some pointers might be 0
	  //TrackCandidateVec is cleared directly after this loop
      }
      //Don't remove this clear after the loop:
      TrackCandidateVec.clear();
      
      if( !outputCol->empty() ){
	  evt->addCollection( outputCol, _outputColName );
      }else{
	  delete outputCol;
      }
      
      streamlog_out(DEBUG0)<<"evt->addCollection "<<std::endl;
      
      
      streamlog_out(DEBUG0)<<"end process event "<<evt->getEventNumber()<<std::endl;
      
      
    }
  
    
  void TrackFinderHoughTrafoProcessor::check( LCEvent * evt ) { 
    streamlog_out(DEBUG0)<<"TrackFinderHoughTrafoProcessor::check( evt "
			 <<evt->getEventNumber()<<")"<<std::endl;
    
    LCCollection* inputCol=0;
    try{      
      inputCol = evt->getCollection( _inputColName ) ;
     }
    
    catch(DataNotAvailableException){
      // if there is not input collection give a warning to the user
      streamlog_out(WARNING2)
	<<"check: input collection "
	<<_inputColName<<" cannot be found in event "<<evt->getEventNumber()<<std::endl;
      return;
    }
    
    
    LCCollection* outputCol=0;
    try{      
      outputCol = evt->getCollection( _outputColName ) ;
    }
    
    catch(DataNotAvailableException){
      // if there is not output collection give a warning to the user
      streamlog_out(WARNING2)
	<<"check: output collection "
	<<_outputColName<<" cannot be found in event "<<evt->getEventNumber()<<std::endl;
      return;
    }
    
    

    std::stringstream sEvt;
    sEvt << evt->getEventNumber();
     
    std::string histoname2 = "HistoEvent" + sEvt.str();
    
    streamlog_out(DEBUG0)<<"creating histo : "<<histoname2<<std::endl;

    AIDAProcessor::histogramFactory(this)->destroy( HistoEvent );
    HistoEvent = AIDAProcessor::histogramFactory(this)->
      createHistogram2D( histoname2.c_str(), "Event in mm", 
			 256 , 0, 14.08, 256 , 0, 14.08 );
    
    
    
    for(int iHit=0; iHit < inputCol->getNumberOfElements() ; iHit++){
      
      streamlog_out(DEBUG0)<<iHit+1<<" von "<<inputCol->getNumberOfElements()<<std::endl;
      
      TrackerHit* RawHit = 
	dynamic_cast<TrackerHit*>( inputCol->getElementAt( iHit ) ) ;
      
      streamlog_out(DEBUG0)<<"hinter RawHit  "<<RawHit<<std::endl;
      
      if( RawHit != 0 ) {

	streamlog_out(DEBUG0)<<iHit+1<<" von "
			     <<inputCol->getNumberOfElements()<<std::endl;	 
	const double *pos = RawHit->getPosition(); //hit position x, y, z in mm
	
	streamlog_out(DEBUG0)<<"HistoEvent->fill("<< pos[0]<<", "<<pos[1]<<") "
			     <<iHit+1<<" von "
			      <<inputCol->getNumberOfElements()<<std::endl;

	
	
	HistoEvent->fill( pos[0], pos[1] );
	
	
      }

      
    }// end for(int iHit=0; iHit < inputColVec.size() ; iHit+
    
    for ( std::vector<AIDA::IHistogram2D*>::iterator i = HistoLineVec.begin();
	  i < HistoLineVec.end(); i++ ) {
	
      AIDAProcessor::histogramFactory(this)->destroy( *i );	
      
    }
    for ( std::vector<AIDA::IHistogram2D*>::iterator i = HistoTrackVec.begin();
	  i < HistoTrackVec.end(); i++ ) {
	
      AIDAProcessor::histogramFactory(this)->destroy( *i );	
      
    }
    
     HistoLineVec.clear();
     HistoTrackVec.clear();
     
     
     for(int iTrack=0; iTrack < outputCol->getNumberOfElements() ; iTrack++){

       streamlog_out(DEBUG0)<<"iTrack = "<<iTrack<<std::endl;

       std::stringstream siTrack;
       siTrack << iTrack;

       std::string histoname3 = "HistoLine" + sEvt.str() + "_Track" + siTrack.str();
       std::string histoname4 = "HistoTrack" + sEvt.str() + "_Track" + siTrack.str();

       streamlog_out(DEBUG0)<<"creating histo : "<<histoname3<<std::endl;

       HistoLineVec.push_back( AIDAProcessor::histogramFactory(this)->
			       createHistogram2D( histoname3.c_str(), "track in mm", 
						  256 , 0, 14.08, 256 , 0, 14.08 ) );

       AIDA::IHistogram2D *HistoLine = *(HistoLineVec.end()-1);


       streamlog_out(DEBUG0)<<"creating histo : "<<histoname4<<std::endl;

       HistoTrackVec.push_back( AIDAProcessor::histogramFactory(this)->
			       createHistogram2D( histoname4.c_str(), "track in mm", 
						  256 , 0, 14.08, 256 , 0, 14.08 ) );

       AIDA::IHistogram2D *HistoTrack = *(HistoTrackVec.end()-1);


       Track* RawTrack = 
	 dynamic_cast<Track*>( outputCol->getElementAt( iTrack ) ) ;
       
       if( RawTrack != 0 ) {


	 std::vector< TrackerHit * > HitVec = RawTrack->getTrackerHits();
	 for ( std::vector< TrackerHit * >::iterator hitIter = HitVec.begin();
	       hitIter < HitVec.end(); hitIter ++ ) {
	   
	   const double *pos = (*hitIter)->getPosition();
	   HistoTrack->fill( pos[0], pos[1] );

	 }


	 // just for debugging: the straight line on one (single) Timepix
	 double pitch = 0.055;
      	 for ( int i = 0; i < 256; i++ ) {
	   
	   HistoLine->fill( i*pitch, 
			    tan( RawTrack->getPhi() )*i*pitch  + RawTrack->getD0() );
	   HistoLine->fill( 1/tan( RawTrack->getPhi() ) *i*pitch  - 
			    RawTrack->getD0()/ tan( ( RawTrack->getPhi() ) )  , 
			    i*pitch );
	   
	 }

       } //end if( RawTrack != 0 )

     }//end for(int iTrack=0; iTrack < outputCol->getNumberOfElements()
	 


     streamlog_out(DEBUG0)<<"end check( evt "<<evt->getEventNumber()<<")"<<std::endl;
    }
    
    
    void TrackFinderHoughTrafoProcessor::end(){ 

      streamlog_out(DEBUG0)<<"TrackFinderHoughTrafoProcessor::end()"<<std::endl;
      /*      	
      streamlog_out(MESSAGE4) << "TrackFinderHoughTrafoProcessor::end()  " << name() 
		<< " processed " << evt->getEventNumber() << " events in " 
		<< _nRun << " runs "
		<< std::endl ;
      */
      for ( std::vector<AIDA::IHistogram2D*>::iterator i = HistoLineVec.begin();
	    i < HistoLineVec.end(); i++ ) {
	  
	  AIDAProcessor::histogramFactory(this)->destroy( *i );	
	  
      }
      for ( std::vector<AIDA::IHistogram2D*>::iterator i = HistoTrackVec.begin();
	    i < HistoTrackVec.end(); i++ ) {
	  
	  AIDAProcessor::histogramFactory(this)->destroy( *i );	
	  
      }
      for ( std::vector<AIDA::IHistogram2D*>::iterator i = HistoHoughVec.begin();
	    i < HistoHoughVec.end(); i++ ) {
	
	AIDAProcessor::histogramFactory(this)->destroy( *i );	

      }

      HistoLineVec.clear();
      HistoTrackVec.clear();
      HistoHoughVec.clear();

    }


  TrackFinderHoughTrafoProcessor::TrackProperties TrackFinderHoughTrafoProcessor::maxBinFinder( AIDA::IHistogram2D * histo ) {

    int xBins = (histo->xAxis()).bins();
    int yBins = (histo->yAxis()).bins();

    double max = 0;
    int xMax = 0;
    int yMax = 0;

    for ( int iterX = 0; iterX < xBins; iterX++) {

      for ( int iterY = 0; iterY< yBins; iterY++) {

	
	double binheight = histo->binHeight( iterX, iterY );

	if ( binheight > max ) { 
	 
	  max = binheight;
	  xMax = iterX;
	  yMax = iterY;

	}

      }

    }

    streamlog_out(DEBUG0)<<"in maxBinFinder:\n"
			 <<"Number of Bins : ("<<xBins<<", "<<yBins<<")\n"
			 <<"Max Bin : ("<<xMax<<", "<<yMax<<"); Height = "<<max
			 <<std::endl;

    streamlog_out(DEBUG0)<<"max x = "<<(histo->xAxis()).binLowerEdge(xMax)
			 <<"  max y = "<<(histo->yAxis()).binLowerEdge(yMax)
			   <<std::endl; 

    double x = ( (histo->xAxis()).binUpperEdge( xMax ) + 
                 (histo->xAxis()).binLowerEdge( xMax ) ) /2;

    double y = ( (histo->yAxis()).binUpperEdge( yMax ) + 
                 (histo->yAxis()).binLowerEdge( yMax ) ) /2;


    return TrackProperties( x, y, -1, -1);
    
  }


  double TrackFinderHoughTrafoProcessor::distanceToTrackFinder( const double *pos, 
							     DoubleVec par, 
							     std::string _trackType  ){    
    double d = 0;
    
    if ( _trackType == "StraightLine" ) {

      //d = ( (pos[1] - par[1])/par[0] - pos[0] ) / par[0];
      d = ( pos[1] - (par[0]*pos[0] + par[1]) )*cos( atan( par[0] ) );

    } //StraightLine

    return d;

  }

 

} // end of namespace marlintpc{
