
// stl
#include "stdio.h"
#include <string>

// LCIO
#include "lcio.h"
#include "EVENT/LCIO.h"
#include "IO/LCWriter.h"
#include "IMPL/LCGenericObjectImpl.h"
#include "IMPL/LCEventImpl.h"
#include "IMPL/LCCollectionVec.h"

// marlintpc
#include "FieldMapReader.h"

namespace marlintpc {

FieldMapReader* FieldMapReader::defaultInstance = 0;

FieldMapReader::FieldMapReader() {

}

FieldMapReader::~FieldMapReader() {
	delete defaultInstance;
	delete _file;
}

FieldMapReader* FieldMapReader::READER() {
	if (defaultInstance == 0)
		defaultInstance = new FieldMapReader();
	return defaultInstance;
}

EVENT::LCCollection* FieldMapReader::read_field_map( const std::string filePath ) {

	_field_map = new IMPL::LCCollectionVec( lcio::LCIO::LCGENERICOBJECT );

	_file = new std::ifstream( filePath.c_str() );

	_field_type = "";

	if (_file->is_open()) {

		std::string line;

		while (!_file->eof()) {

			std::getline( *_file, line );

			// skip empty or comment lines
			if ((line != "") && (line.substr(0,2) != "//")) {

				if (line == "### Start Field Map ###") {

					if (_field_type == "Cylindrical2DMagneticField") {
						std::cout << "Reading Cylindrical2DMagneticField from " << filePath << " ... " << std::endl;
						read_Cylindrical2DMagneticField();
					} else if (_field_type == "Cylindrical3DMagneticField") {
						std::cout << "Reading Cylindrical3DMagneticField from " << filePath << " ... " << std::endl;
						read_Cylindrical3DMagneticField();
					} else {
						std::cerr << filePath << " contains no known type of field." << std::endl;
						delete _field_map;
						return 0;
					}

				} else {
					read_parameter( line );
				}

			}

		}

	} else {
		std::cerr << "Unable to open " << filePath << std::endl;
		delete _field_map;
		_file->close();
		return 0;
	}

	_file->close();
	return _field_map;
}

void FieldMapReader::set_verbose( const int verbose_level ) {
	_verbose = verbose_level;
}

void FieldMapReader::read_Cylindrical2DMagneticField() {

	std::string(line);

	while (!_file->eof()) {

		std::getline( *_file, line );

		// skip empty or comment lines
		if ((line != "") && (line.substr(0,2) != "//")) {

			float r = 0.;
			float z = 0.;
			float Br = 0.;
			float Bz = 0.;
			sscanf( line.c_str(), "%f %f %f %f", &r, &z, &Br, &Bz );

			if (_verbose > 1)
				std::cout << "r: " << r << ", z: " << z
						  << ", Br: " << Br << ", Bz: " << Bz << std::endl;

			IMPL::LCGenericObjectImpl* field_map_entry = new IMPL::LCGenericObjectImpl(0, 2, 0);
			field_map_entry->setFloatVal(0, Br);
			field_map_entry->setFloatVal(1, Bz);
			_field_map->addElement(field_map_entry);
		}

	}

}

void FieldMapReader::read_Cylindrical3DMagneticField() {

	std::string(line);

	while (!_file->eof()) {

		std::getline( *_file, line );

		// skip empty or comment lines
		if ((line != "") && (line.substr(0,2) != "//")) {

			float r = 0.;
			float phi = 0.;
			float z = 0.;
			float Br = 0.;
			float Bphi = 0.;
			float Bz = 0.;
			sscanf( line.c_str(), "%f %f %f %f %f %f", &r, &phi, &z, &Br, &Bphi, &Bz );

			if (_verbose > 1)
				std::cout << "r: " << r << ", phi: " << phi << ", z: " << z
						  << ", Br: " << Br << ", Bphi: " << Bphi << ", Bz: " << Bz << std::endl;

			IMPL::LCGenericObjectImpl* field_map_entry = new IMPL::LCGenericObjectImpl(0, 3, 0);
			field_map_entry->setFloatVal(0, Br);
			field_map_entry->setFloatVal(1, Bphi);
			field_map_entry->setFloatVal(2, Bz);
			_field_map->addElement(field_map_entry);
		}

	}

}

void FieldMapReader::read_parameter( std::string line ) {

	char c_type [100];
	char c_parameter [1000];
	sscanf( line.c_str(), "%s %s", c_type, c_parameter );
	std::string type = c_type;
	std::string parameter = c_parameter;

	if (type == "FieldType") {
		if (_field_type != "") {
			if (_verbose > 0)
				std::cerr << "Warning: multiple definitions of FieldType" << std::endl;
		}
		_field_map->parameters().setValue(type, parameter);
		_field_type = parameter;
	} else if (type == "String") {
		size_t length_of_first_part = type.length() + 1 + parameter.length() + 1;
		std::string value = line.substr(length_of_first_part, line.length() - length_of_first_part);
		_field_map->parameters().setValue(parameter, value);
		if (_verbose > 0)
			std::cout << "Reading string - " << parameter << ": "  << value << std::endl;
	} else if (type == "Int") {
		int value = 0;
		sscanf( line.c_str(), "%*s %*s %i", &value );
		_field_map->parameters().setValue(parameter, value);
		if (_verbose > 0)
			std::cout << "Reading int - " << parameter << ": " << value << std::endl;
	} else if (type == "Float") {
		float value = 0.;
		sscanf( line.c_str(), "%*s %*s %f", &value );
		_field_map->parameters().setValue(parameter, value);
		if (_verbose > 0)
			std::cout << "Reading float - " << parameter << ": " << value << std::endl;
	} else {
		if (_verbose > 0)
			std::cerr << "Unknown parameter type \"" << type << "\". Ignoring line \"" << line << "\"" << std::endl;
	}

}

} // namespace marlintpc
