
// stl
#include <string>

// marlintpc
#include "MagneticField.h"


namespace marlintpc {

/** constructor
 */
MagneticField::MagneticField( std::string field_typename ):
	Field(field_typename) {
	set_translation(CLHEP::Hep3Vector());
	set_rotation( 0.0, 0.0 );
	set_scale_factor(1.0);
}


/** destructor
 */
MagneticField::~MagneticField() {}


/** sets the relative position of the origin of the field map (i.e. center of the coil) to the general coordinate system.
 * The translation is applied after the rotation, so when the axis of original system and coil system are parallel.
 */
void MagneticField::set_translation( const CLHEP::Hep3Vector &vector ) {
	_translation = vector;
}

/** sets the rotations if the coordinate system of the field map is tilted with respect to the general coordinate system.
 * The rotations are applied before a possible translation.
 * @param x_rotation The angle between the two z axis in y direction (in rad)
 * @param y_rotation The angle between the two z axis in x direction (in rad)
 */
void MagneticField::set_rotation( const double &x_rotation, const double &y_rotation ) {
	_x_rotation = x_rotation;
	_y_rotation = y_rotation;
}

/** sets the scale factor for the magnetic field, i.e. higher current than during mapping
 */
void MagneticField::set_scale_factor( const double &scale_factor ) {
	_scale_factor = scale_factor;
}

/** gets the field value at a specified point
 */
CLHEP::Hep3Vector MagneticField::get_field_value( const CLHEP::Hep3Vector &position ) {

	CLHEP::Hep3Vector _position = position;
	_position.rotateX( _x_rotation ).rotateY( _y_rotation );	// rotate into coil coordinate system
	_position += -_translation;	// translate into coil coordinate system

	CLHEP::Hep3Vector field_value = get_field_value_coil_coordinates( _position );

	field_value = field_value.rotateY( -_y_rotation).rotateX( -_x_rotation );	// rotate back into TPC coordinate system

	return field_value*_scale_factor;
}

/** gets the field error. Assumes that the error is symmetric and independent of position
 */
CLHEP::Hep3Vector MagneticField::get_field_error( ) {

	// need to rotate into TPC coordinate system
	return _field_error.rotateY( -_y_rotation ).rotateX( -_x_rotation );
}


} // namespace marlintpc
