//--------------------------------------------------------------------------
// File and Version Information:
// 	$Id: HepRepXMLWriter.cc,v 1.5 2002/02/05 12:04:25 perl Exp $
//
// Description:
//	Utility for the creation of HepRep XML Files (HepRep version 1).
//
//      For details, see:
//      http://www.slac.stanford.edu/~perl/HepRepXMLWriter.html
//
// Environment:
//	Software developed for the general High Energy Physics community.
//
// Author :
//      Joseph Perl                 Original Author
//
// Copyright Information:
//      Copyright (C) 2002          Stanford Linear Accelerator Center
//------------------------------------------------------------------------
#include <fstream>
#include <unistd.h>
#include <iostream>
#include <cstring>

#include "HepRepXMLWriter.h"

using namespace std;

HepRepXMLWriter::HepRepXMLWriter()
{
  isOpen = false;
  init();
}

void HepRepXMLWriter::init()
{
  typeDepth = -1;

  int i = -1;
  while (i++<49) {
    prevTypeName[i] = new char[1];
    strcpy(prevTypeName[i],"");

    inType[i] = false;
    inInstance[i] = false;
  }

  inPrimitive = false;
  inPoint = false;
}

void HepRepXMLWriter::addType(const char* name,int newTypeDepth)
{
  if (fout.good())
  {
    // Flatten structure if it exceeds maximum allowed typeDepth of 49.
    if (newTypeDepth > 49)
      newTypeDepth = 49;

    if (newTypeDepth < 0)
      newTypeDepth = 0;

    // Insert any layers that are missing from the hierarchy (protects against
    // callers that skip from, say, layer 1 to layer 3 with no layer 2).
    while (typeDepth < (newTypeDepth-1)) {
      addType("Layer Inserted by HepRepXMLWriter", typeDepth + 1);
      addInstance();
    }

    // If moving closer to the root, close previously open types.
    while (newTypeDepth<typeDepth)
      endType();

    // Close any remaining primitives of the current instance.
    endPrimitive();

    // If this is a new type name for the current depth, declare the
    // new Type.  Otherwise, it is just another Instance of the current Type.
    if (strcmp(name,prevTypeName[newTypeDepth])!=0)
    {
      if (inType[newTypeDepth])
	endType();

      prevTypeName[newTypeDepth] = new char[strlen(name)+1];
      strcpy(prevTypeName[newTypeDepth],name);

      inType[newTypeDepth] = true;
      indent();
      fout << "<heprep:type version=\"null\" name=\"" << name << "\">"
	 << std::endl;

      typeDepth = newTypeDepth;
    }
  } else {
    cout << "HepRepXMLWriter:addType No file is currently open." << endl;
  }
}

void HepRepXMLWriter::addInstance()
{
  if (fout.good())
  {
    if (inType[typeDepth])
    {
      endInstance();
      inInstance[typeDepth] = true;
      indent();
      fout << "<heprep:instance>" << endl;
    } else {
      cout <<
	"HepRepXMLWriter:addInstance No HepRep Type is currently open"
	     << endl;
    }
  } else {
    cout << "HepRepXMLWriter:addInstance No file is currently open"
	   << endl;
  }
}

void HepRepXMLWriter::addPrimitive()
{
  if (fout.good())
  {
    if (inInstance[typeDepth])
    {
      endPrimitive();
      inPrimitive = true;
      indent();
      fout << "<heprep:primitive>" << endl;
    } else {
      cout <<
	"HepRepXMLWriter:addPrimitive No HepRep Instance is currently open"
	     << endl;
    }
  } else {
    std::cout << "HepRepXMLWriter:addPrimitive No file is currently open"
	   << endl;
  }
}

void HepRepXMLWriter::addPoint(double x, double y, double z)
{
  if (fout.good())
  {
    if (inPrimitive)
    {
      endPoint();
      inPoint = true;
      indent();
      fout << "<heprep:point x=\"" << x << "\" y=\"" << y << "\" z=\"" << z << "\">" << endl;
    } else {
      cout <<
	"HepRepXMLWriter:addPoint No HepRep Primitive is currently open"
	     << endl;
    }
  } else {
    cout << "HepRepXMLWriter:addPoint No file is currently open" << endl;
  }
}

void HepRepXMLWriter::addAttDef(const char* name,
	       const char* desc,
	       const char* type,
	       const char* extra)
{
  if (fout.good())
  {
    indent();
    fout << "<heprep:attdef extra=\"" << extra << "\" name=\"" << name << "\" type=\"" << type << "\"" << endl;
    indent();
    fout << "  desc=\"" << desc << "\"/>" << endl;
  } else {
    cout << "HepRepXMLWriter:addAttDef No file is currently open" << endl;
  }
}

// Four methods to fill attValues
void HepRepXMLWriter::addAttValue (const char* name,
		  const char* value)
{
  if (fout.good())
  {
    indent();
    fout << "  <heprep:attvalue showLabel=\"NONE\" name=\"" << name << "\"" << endl;
    indent();
    fout << "    value=\"" << value << "\"/>" << endl;
  } else {
    cout << "HepRepXMLWriter:addAttValue No file is currently open" << endl;
  }
}

void HepRepXMLWriter::addAttValue (const char* name,
		  double value)
{
  if (fout.good())
  {
    indent();
    fout << "  <heprep:attvalue showLabel=\"NONE\" name=\"" << name << "\"" << endl;
    indent();
    fout << "    value=\"" << value << "\"/>" << endl;
  } else {
    cout << "HepRepXMLWriter:addAttValue No file is currently open" << endl;
  }
}

void HepRepXMLWriter::addAttValue (const char* name,
		  int value)
{
  if (fout.good())
  {
    indent();
    fout << "  <heprep:attvalue showLabel=\"NONE\" name=\"" << name << "\"" << endl;
    indent();
    fout << "    value=\"" << value << "\"/>" << endl;
  } else {
    cout << "HepRepXMLWriter:addAttValue No file is currently open" << endl;
  }
}

void HepRepXMLWriter::addAttValue (const char* name,
		  bool value)
{
  if (fout.good())
  {
    indent();
    fout << "  <heprep:attvalue showLabel=\"NONE\" name=\"" << name << "\"" << endl;
    indent();
    if (value)
      fout << "    value=\"True\"/>" << endl;
    else
      fout << "    value=\"False\"/>" << endl;
  } else {
    cout << "HepRepXMLWriter:addAttValue No file is currently open" << endl;
  }
}

void HepRepXMLWriter::addAttValue (const char* name,
				   double value1,
				   double value2,
				   double value3)
{
  if (fout.good())
  {
    int redness = int(value1*255.);
    int greenness = int(value2*255.);
    int blueness = int(value3*255.);
    indent();
    fout << "  <heprep:attvalue showLabel=\"NONE\" name=\"" << name << "\"" << endl;
    indent();
    fout << "    value=\"" << redness << "," << greenness << "," << blueness << "\"/>" << endl;
  } else {
    cout << "HepRepXMLWriter:addAttValue No file is currently open" << endl;
  }
}

void HepRepXMLWriter::open(const char* fileSpec)
{
  if (isOpen)
    close();
  
  fout.open(fileSpec);
    
  if (fout.good()) {
    fout << "<?xml version=\"1.0\" ?>" << endl;
    fout << "<heprep:heprep xmlns:heprep=\"http://www.freehep.org/HepRep\"" << endl;
    fout << "  xmlns:xsi=\"http://www.w3.org/1999/XMLSchema-instance\" xsi:schemaLocation=\"HepRep.xsd\">" << endl;
    
    isOpen = true;
    init();
  } else {
    cout << "HepRepXMLWriter:open Unable to write to file " << fileSpec << endl;
  }
}

void HepRepXMLWriter::close()
{
  // Close any remaining open Types
  endTypes();

  if (fout.good()) {      
    fout << "</heprep:heprep>" << endl;
    fout.close( );
    isOpen = false;
  } else {
    cout << "HepRepXMLWriter:close No file is currently open" << endl;
  }
}

void HepRepXMLWriter::endTypes()
{
  // Close any remaining open Types
    while(typeDepth>-1)
      endType();
}

void HepRepXMLWriter::endType()
{
  endInstance();
  indent();
  fout << "</heprep:type>" << endl;
  inType[typeDepth] = false;
  delete [] prevTypeName[typeDepth];
  prevTypeName[typeDepth] = new char[1];
  strcpy(prevTypeName[typeDepth],"");
  typeDepth--;
}

void HepRepXMLWriter::endInstance()
{
  if (inInstance[typeDepth])
  {
    endPrimitive();
    indent();
    fout << "</heprep:instance>" << endl;
    inInstance[typeDepth] = false;
  }
}

void HepRepXMLWriter::endPrimitive()
{
  if (inPrimitive)
  {
    endPoint();
    indent();
    fout << "</heprep:primitive>" << endl;
    inPrimitive = false;
  }
}

void HepRepXMLWriter::endPoint()
{
  if (inPoint)
  {
    indent();
    fout << "</heprep:point>" << endl;
    inPoint = false;
  }
}

void HepRepXMLWriter::indent()
{
  if (fout.good())
  {
    int i = 0;
    while (inType[i] && i<12) {
      fout << "  ";
      if (inInstance[i])
        fout << "  ";
      i++;
    }

    if (inPrimitive)
      fout << "  ";
    if (inPoint)
      fout << "  ";
  }
}
