#include "TimePixParameters.h"

// the data format is as follow:
// there are 4 leading integers:
// 0 : chipID
// 1 : type
// 2 : polarity
// 3 : nChips
// then the block of nChips*16 DAC values follows
// if the data format is extended only add trailing integers to 
// preserve backward compatibility
#define N_LEADING_INTEGERS 4
#define N_TRAILING_INTEGERS 0


namespace tpcconddata
{

  TimePixParameters::TimePixParameters(int nChips)
  {
    createObj = true;
    myObj     = new IMPL::LCGenericObjectImpl(N_LEADING_INTEGERS + nChips*16 + N_TRAILING_INTEGERS,
					      0, // no floats
					      2);// two doubles

    myObj->setIntVal(3, nChips);
    setType(3);   // set the default type to TimePix
  }

  TimePixParameters::TimePixParameters(EVENT::LCObject* obj)
  {
    createObj=false;
    myObj = dynamic_cast<IMPL::LCGenericObjectImpl*>( obj );
    
    if (myObj ==0){
	//could be a TimePixParameters object
	TimePixParameters* dummyObj = dynamic_cast<TimePixParameters*>( obj );
	if (dummyObj != 0){
	    myObj=dummyObj->myObj;
	}
    }
  }
    
  TimePixParameters::~TimePixParameters()
  {
    if (createObj) delete myObj;
  }

  void TimePixParameters::setChipID (int chipID){ myObj->setIntVal(0, chipID); }
  int  TimePixParameters::getChipID() const{ return myObj->getIntVal(0); }
    
  void TimePixParameters::setDACs(int chipIndex, int * DACs)
  {
    for (int i=0; i<16; i++)
    {
      myObj->setIntVal( N_LEADING_INTEGERS + chipIndex*16 + i , DACs[i] );
    }
  }
  //  const int * TimePixParameters::getDACs(int chipIndex) const
  //{
  // OK, this does not work since the getIntVal gives a copy, not a reference :-(
  //  return  &myObj->getIntVal( getChipID() + chipIndex*16 );
  //}

  
    /// Get a specific DAC by index
  int TimePixParameters::getDAC(int chipIndex, int dacIndex) const
  {
    return  myObj->getIntVal(N_LEADING_INTEGERS + chipIndex*16 + dacIndex);
  }
  
  void TimePixParameters::setType(int type) { myObj->setIntVal(1, type); }
    /// get the chip type (3=timepix) 
  int TimePixParameters::getType() const { return  myObj->getIntVal(1); }
    
    /// set the polarity
    void TimePixParameters::setPolarity(int polarity){ myObj->setIntVal(2, polarity); }
    /// get the polarity
    int TimePixParameters::getPolarity() const { return  myObj->getIntVal(2); }
    
    // get the DACs for the Timepix by name (convenience functions for not to use the index)
    int TimePixParameters::getIKrumm(int chipIndex) const   {return getDAC( chipIndex,  0);}
    int TimePixParameters::getDisc(int chipIndex) const     {return getDAC( chipIndex,  1);}
    int TimePixParameters::getPreamp(int chipIndex) const   {return getDAC( chipIndex,  2);}
    int TimePixParameters::getBuffA(int chipIndex) const    {return getDAC( chipIndex,  3);}
    int TimePixParameters::getBuffB(int chipIndex) const    {return getDAC( chipIndex,  4);}
    int TimePixParameters::getDelayN(int chipIndex) const   {return getDAC( chipIndex,  5);}
    int TimePixParameters::getTHLFine(int chipIndex) const  {return getDAC( chipIndex,  6);}
    int TimePixParameters::getTHLCoarse(int chipIndex) const{return getDAC( chipIndex,  7);}
    int TimePixParameters::getTHHFine(int chipIndex) const  {return getDAC( chipIndex,  8);}
    int TimePixParameters::getTHHCoarse(int chipIndex) const{return getDAC( chipIndex,  9);} 
    int TimePixParameters::getFBK(int chipIndex) const 	    {return getDAC( chipIndex, 10);}
    int TimePixParameters::getGnd(int chipIndex) const 	    {return getDAC( chipIndex, 11);}
    int TimePixParameters::getTHS(int chipIndex) const 	    {return getDAC( chipIndex, 12);}
    int TimePixParameters::getBiasVLDS(int chipIndex) const {return getDAC( chipIndex, 13);}
    int TimePixParameters::getRefLVDS(int chipIndex) const  {return getDAC( chipIndex, 14);} 
    
    /// Set the clock (in MHz)
    void TimePixParameters::setClock(double clock) { myObj->setDoubleVal(0, clock); }
    /// Get the clock (in MHz)
    double TimePixParameters::getClock() const { return  myObj->getDoubleVal(0); }
    
    /// Set the shutter end time (wrt. tigger)
    void TimePixParameters::setShutterEndTime(double shutterEndTime){ myObj->setDoubleVal(1, shutterEndTime); }
    /// Get  the shutter end time (wrt. tigger)
    double TimePixParameters::getShutterEndTime() const{ return  myObj->getDoubleVal(1); }

    /// Get the number of chips
    int TimePixParameters::getNChips() const  { return  myObj->getIntVal(3); }
    
  std::string TimePixParameters::getRevision() {return std::string("$Rev: $");}
  std::string TimePixParameters::getDefaultColName() 
  {
    return std::string("TimePixParameters");
  }
    
  //    void TimePixParameters::print(  std::ostream& os ) const;
    
    // -------- need to implement abstract methods from LCGenericObject
  int TimePixParameters::getNInt() const { return  N_LEADING_INTEGERS
                                                  + getNChips()*16 
                                                  + N_TRAILING_INTEGERS;}
  int TimePixParameters::getNFloat() const {return 0; } 
  int TimePixParameters::getNDouble() const { return 2; }
    
  int TimePixParameters::getIntVal(int index) const {return myObj->getIntVal( index);}
    float TimePixParameters::getFloatVal(int index) const{return myObj->getFloatVal( index);}
    double TimePixParameters::getDoubleVal(int index) const{return myObj->getDoubleVal( index);}

  void TimePixParameters::print ( std::ostream& os ) const {
    os << "BoardID " << getChipID() << " , Type "<< getType() << " , " << getNChips() << " Chips"
       <<std::endl;
    for ( int chip=0; chip<getNChips() ; chip++)
    {
      os << "Chip " << chip << "DACs ";
      for ( int dacIndex=0; dacIndex<16 ; dacIndex++)
      {
	os << "\t" << getDAC(chip, dacIndex) ;
      }
      os << std::endl;
    }
  } 
  
  std::ostream &operator<<(std::ostream &os, const TimePixParameters &p) {
    p.print(os);
    return os;
  }
}                                //end namespace tpcconddata

