#ifndef ADCRawEventDisplayProcessor_h
#define ADCRawEventDisplayProcessor_h 1

// LCIO
#include "lcio.h"

// Marlin
#include "marlin/Processor.h"

namespace AIDA{
    class IHistogram1D;  
}

namespace marlintpc{

    /**  This processor creates one histogram with the raw data of each channel.
     *   This means for n channels there are n histograms per event!!!
     *   Be careful only to use it for dedicated debugging runs.
     *
     *   Although it is called RAW event processor it does not use TrackerRawData but TrackerData,
     *   so it can also be used for calibrated (e. g. pedestal subtracted) data. Simply use the
     *   TrackerRawDataToDataConverterProcessor before if you have TrackerRawData.
     * 
     *
     *  <h4>Input - Prerequisites</h4>
     *  @param InputCollectionName Name of the \c TrackerData input collection  (default TPCData)
     *  @param nTimeSamples Number of time samples (bins per histogram). Default: 1024
     *  @param deleteHistograms Set true if running with zero suppressed data. Otherwise histograms will only be reset (emptied). Default: false
     *
     *  <h4>Output</h4> 
     *  One histogram per channel and event
     * 
     * @author Martin Killenber (University of Bonn)
     */
    class ADCRawEventDisplayProcessor : public marlin::Processor {
	
    public:
	
	/* This method will be called by the marlin package
	 * It returns a processor of the currend type (here AltroRawEventDisplayProcessor)
	 */
	virtual Processor*  newProcessor() { return new ADCRawEventDisplayProcessor ; }
	
	/* the default constructor
	 * here the processor parameters are registered to the marlin package
	 * other initialisation should be placed in the init method
	 */
	ADCRawEventDisplayProcessor() ;
	
	/* Called at the beginning of the job before anything is read.
	 * Use to initialize the processor, e.g. book histograms
	 */
	virtual void init() ;
	
	/* Called for every run.
	 * in this processor it is not used
	 */
	virtual void processRunHeader( lcio::LCRunHeader* run ) ;
	
	/* Called for every event - the working horse.
	 * Here the real conversion take place
	 */
	virtual void processEvent( lcio::LCEvent * evt ) ; 
	
	/* This method is only called if the check flag is set (default) in the main processor
	 * (where the input files are given)
	 * It should be used to create check plots which are not needed in a complete analysis
	 * (where normaly the check flag is deactivated)
	 */
	virtual void check( lcio::LCEvent * evt ) ; 
	
	
	/* Called after data processing for clean up.
	 * e.g. saving control histogramms, delete created objects
	 * (which are not stored in a collection/event), etc.
	 */
	virtual void end() ;
	
	
    protected:
	
	/* normally the doxygen documentation is palced before the method/variable
	 * but for one line comments in case of variables it may be more convenied for you
	 * to place it after the variable. In this case you use an arrow to tell doxygen
	 * to which variable the commet belongs to.
	 */
	int _nTimeSamples; ///< Number of time samples (bins per histogram)
	std::string _inputColName; ///< Name of the input collection
	
	///< Set true if running with zero suppred data. Otherwise histograms will only be reset (emptied)
	bool _deleteHistograms;

	std::map<std::pair<int,int>, AIDA::IHistogram1D*> rawHistos;

    } ;
    
}

#endif
    
    

